import pytest
import os
import subprocess
import time
import json
import yaml
import xarray as xr
import tomato
import zmq

from . import utils

PORT = 12345
CTXT = zmq.Context()


@pytest.mark.parametrize(
    "casename, npoints, prefix",
    [
        ("counter_1_0.1", 10, "results.1"),
        ("counter_5_0.2", 25, "results.1"),
        ("counter_output_prefix", 20, "data"),
        ("counter_output_path", 20, os.path.join("newfolder", "results.1")),
        ("counter_multistep", 15, "results.1"),
    ],
)
def test_counter_npoints_metadata(
    casename, npoints, prefix, datadir, start_tomato_daemon, stop_tomato_daemon
):
    os.chdir(datadir)
    utils.run_casenames([casename], [None], ["pip-counter"])
    status = utils.job_status(1)
    while status in {"q", "qw", "r"}:
        time.sleep(1)
        status = utils.job_status(1)
    assert status == "c"
    files = os.listdir(os.path.join(".", "Jobs", "1"))
    assert "jobdata.json" in files
    assert "job-1.log" in files
    if prefix is not None:
        assert os.path.exists(f"{prefix}.nc")
        dt = xr.open_datatree(f"{prefix}.nc")
        assert "tomato_version" in dt.attrs
        assert "tomato_Job" in dt.attrs

        ds = dt["counter"]
        print(f"{ds=}")
        assert ds["uts"].size == npoints
        assert "tomato_Component" in ds.attrs


@pytest.mark.parametrize(
    "casename",
    [
        "counter_60_0.1",
    ],
)
def test_counter_cancel(casename, datadir, start_tomato_daemon, stop_tomato_daemon):
    os.chdir(datadir)
    cancel = True
    utils.run_casenames([casename], [None], ["pip-counter"])
    status = utils.job_status(1)
    while status in {"q", "qw", "r", "rd"}:
        time.sleep(2)
        if cancel and status == "r":
            subprocess.run(["ketchup", "cancel", "-p", "12345", "1"])
            cancel = False
            time.sleep(2)
        status = utils.job_status(1)
    assert status == "cd"


@pytest.mark.parametrize(
    "casename,  external",
    [
        ("counter_20_1", True),
        ("counter_snapshot", False),
    ],
)
def test_counter_snapshot_metadata(
    casename, external, datadir, start_tomato_daemon, stop_tomato_daemon
):
    os.chdir(datadir)
    utils.run_casenames([casename], [None], ["pip-counter"])
    time.sleep(5)
    status = utils.job_status(1)
    while status in {"q", "qw", "r"}:
        time.sleep(1)
        if external and status == "r":
            subprocess.run(["ketchup", "snapshot", "-p", "12345", "1"])
            external = False
        status = utils.job_status(1)
    assert status == "c"
    assert os.path.exists("snapshot.1.nc")
    dt = xr.open_datatree("snapshot.1.nc")
    assert "tomato_version" in dt.attrs
    assert "tomato_Job" in dt.attrs
    for group in dt:
        assert "tomato_Component" in dt[group].attrs


@pytest.mark.parametrize(
    "casename, npoints",
    [
        ("counter_multidev", {"counter-1": 10, "counter-2": 5}),
        ("counter_multistep_multidev", {"counter-1": 10, "counter-2": 20}),
    ],
)
def test_counter_multidev(casename, npoints, datadir, stop_tomato_daemon):
    os.chdir(datadir)
    with open("devices_multidev.json", "r") as inf:
        jsdata = json.load(inf)
    with open("devices.yml", "w") as ouf:
        yaml.dump(jsdata, ouf)
    subprocess.run(["tomato", "init", "-p", f"{PORT}", "-A", ".", "-D", ".", "-L", "."])
    subprocess.run(["tomato", "start", "-p", f"{PORT}", "-A", ".", "-vv"])
    utils.wait_until_tomato_running(port=PORT, timeout=3000)

    utils.run_casenames([casename], [None], ["pip-multidev"])
    utils.wait_until_ketchup_status(jobid=1, status="r", port=PORT, timeout=10000)
    utils.wait_until_ketchup_status(jobid=1, status="c", port=PORT, timeout=10000)
    time.sleep(2)

    status = utils.job_status(1)
    assert status == "c"
    files = os.listdir(os.path.join(".", "Jobs", "1"))
    assert "jobdata.json" in files
    assert "job-1.log" in files
    assert os.path.exists("results.1.nc")
    dt = xr.open_datatree("results.1.nc")
    for group, points in npoints.items():
        print(f"{dt[group]=}")
        assert dt[group]["uts"].size == points


def test_counter_measure_task_measure(datadir, start_tomato_daemon, stop_tomato_daemon):
    os.chdir(datadir)
    utils.wait_until_tomato_drivers(port=PORT, timeout=3000)
    kwargs = dict(port=PORT, timeout=1000, context=CTXT)
    ret = tomato.passata.measure(
        name="example_counter:(example-addr,1)",
        **kwargs,
    )
    assert ret.success

    utils.run_casenames(["counter_5_0.2"], [None], ["pip-counter"])
    utils.wait_until_ketchup_status(jobid=1, status="r", port=PORT, timeout=5000)
    ret = tomato.passata.measure(
        name="example_counter:(example-addr,1)",
        **kwargs,
    )
    assert not ret.success
    assert "measurement already running" in ret.msg

    utils.wait_until_ketchup_status(jobid=1, status="c", port=PORT, timeout=5000)
    ret = tomato.passata.measure(
        name="example_counter:(example-addr,1)",
        **kwargs,
    )
    assert ret.success



"""
**tomato.daemon**: module of functions comprising the tomato daemon
-------------------------------------------------------------------
.. codeauthor::
    Peter Kraus

"""

import logging
import argparse
from pathlib import Path
from threading import Thread
import toml
import time
import zmq

from tomato.models import Reply, Daemon
import tomato.daemon.cmd as cmd
import tomato.daemon.job
import tomato.daemon.driver
import tomato.daemon.io as io

logger = logging.getLogger(__name__)


def setup_logging(daemon: Daemon):
    """
    Helper function to set up logging (folder, filename, verbosity, format) based on
    the passed daemon state.
    """
    logdir = Path(daemon.settings["logdir"])
    logdir.mkdir(parents=True, exist_ok=True)
    logfile = logdir / f"daemon_{daemon.port}.log"
    logging.basicConfig(
        level=daemon.verbosity,
        format="%(asctime)s - %(levelname)8s - %(name)-30s - %(message)s",
        handlers=[logging.FileHandler(logfile, mode="a")],
    )


def tomato_daemon():
    """
    The function called when `tomato-daemon` is executed.

    Manages the state of the tomato daemon, including recovery of state via
    :mod:`~tomato.daemon.io`, processing state updates via :mod:`~tomato.daemon.cmd`,
    and the manager threads for both jobs (:mod:`~tomato.daemon.job`) and drivers
    (:mod:`~tomato.daemon.driver`).
    """
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("--port", "-p", type=int, default=1234)
    parser.add_argument("--verbosity", "-V", type=int, default=logging.INFO)
    parser.add_argument("--appdir", "-A", type=str, default=str(Path.cwd()))

    args = parser.parse_args()
    settings = toml.load(Path(args.appdir) / "settings.toml")

    daemon = Daemon(**vars(args), status="bootstrap", settings=settings)
    setup_logging(daemon)
    logger.info("logging set up with verbosity %s", daemon.verbosity)

    logger.debug("attempting to restore daemon state")
    io.load(daemon)

    context = zmq.Context()
    rep = context.socket(zmq.REP)
    logger.debug("binding zmq.REP socket on port %d", daemon.port)
    rep.bind(f"tcp://127.0.0.1:{daemon.port}")
    poller = zmq.Poller()
    poller.register(rep, zmq.POLLIN)

    logger.debug("entering main loop")
    jmgr = Thread(target=tomato.daemon.job.manager, args=(daemon.port,))
    jmgr.do_run = True
    jmgr.start()
    dmgr = Thread(target=tomato.daemon.driver.manager, args=(daemon.port,))
    dmgr.do_run = True
    dmgr.start()
    t0 = time.process_time()
    while True:
        socks = dict(poller.poll(1000))
        if rep in socks:
            msg = rep.recv_pyobj()
            logger.debug(f"received {msg=}")
            if "cmd" not in msg:
                logger.error(f"received msg without cmd: {msg=}")
                ret = Reply(success=False, msg="received msg without cmd", data=msg)
            elif hasattr(cmd, msg["cmd"]):
                ret = getattr(cmd, msg["cmd"])(msg, daemon)
            else:
                logger.error(f"received msg with an invalid cmd: {msg=}")
            logger.debug(f"reply with {ret=}")
            rep.send_pyobj(ret)
        if daemon.status == "stop":
            for mgr, label in [(jmgr, "job"), (dmgr, "driver")]:
                if mgr is not None and mgr.do_run:
                    logger.debug("stopping %s manager thread", label)
                    mgr.do_run = False
            if jmgr is not None:
                jmgr.join(1e-3)
                if not jmgr.is_alive():
                    jmgr = None
                    logger.info("job manager thread joined")
            if dmgr is not None:
                dmgr.join(1e-3)
                if not dmgr.is_alive():
                    dmgr = None
                    logger.info("driver manager thread joined")
            if jmgr is None and dmgr is None:
                io.store(daemon)
                break
        tN = time.process_time()
        if tN - t0 > 10:
            io.store(daemon)
            t0 = tN
    logger.critical("tomato-daemon on port %d is exiting", daemon.port)



"""
**DriverInterface-2.0**
-----------------------
.. codeauthor::
    Peter Kraus


"""

from abc import ABCMeta, abstractmethod
from typing import TypeVar, Any, Union, TypeAlias, Optional
from pydantic import BaseModel, Field
from threading import Thread, current_thread, RLock
from queue import Queue
from tomato.models import Reply
from dgbowl_schemas.tomato.payload import Task
import logging
from functools import wraps
import xarray as xr
import time
import atexit
import pint
import sys

logger = logging.getLogger(__name__)


Type: TypeAlias = type
Val = TypeVar("Val", str, int, float, pint.Quantity)
Key: TypeAlias = tuple[str, str]


def in_devmap(func):
    """
    Helper decorator for coercing ``(address, channel)`` into ``key``
    and checking that it is in :obj:`self.devmap`.
    """

    @wraps(func)
    def wrapper(self, **kwargs):
        if "key" in kwargs:
            key = kwargs.pop("key")
        else:
            address = kwargs.get("address")
            channel = kwargs.get("channel")
            key = (address, channel)
        if key not in self.devmap:
            msg = f"dev with address {address!r} and channel {channel} is unknown"
            return Reply(success=False, msg=msg, data=self.devmap.keys())
        return func(self, **kwargs, key=key)

    return wrapper


def to_reply(func):
    """
    Helper decorator for coercing tuples into :class:`Reply`.
    """

    @wraps(func)
    def wrapper(self, **kwargs):
        ret = func(self, **kwargs)
        if isinstance(ret, Reply):
            return ret
        else:
            success, msg, data = ret
            return Reply(success=success, msg=msg, data=data)

    return wrapper


def log_errors(func):
    """
    Helper decorator for logging all kinds of errors.

    This decorator should be only used on functions in the API of the
    :class:`ModelInterface`, as the caught exceptions will cause the
    driver process to exit.
    """

    @wraps(func)
    def wrapper(self, **kwargs):
        try:
            return func(self, **kwargs)
        except Exception as e:
            logger.critical(e, exc_info=True)
            sys.exit(e)

    return wrapper


class Attr(BaseModel, arbitrary_types_allowed=True):
    """A Pydantic :class:`BaseModel` used to describe device attributes."""

    type: Type
    """Data type of the attribute"""

    rw: bool = False
    """Is the attribute read-write?"""

    status: bool = False
    """Should the attribute be included in component status?"""

    units: str = None
    """Default units for the attribute, optional."""

    maximum: Optional[float | pint.Quantity] = Field(None, union_mode="left_to_right")
    """Maximum value for the attribute, optional."""

    minimum: Optional[float | pint.Quantity] = Field(None, union_mode="left_to_right")
    """Minimum value for the attribute, optional."""


class ModelInterface(metaclass=ABCMeta):
    """
    An abstract base class specifying the driver interface.

    Individual driver modules should expose a :class:`DriverInterface` as a top-level
    import, which inherits from this abstract class. Only the methods of this class
    are used to interact with *drivers* and their *components*.

    This class contains one abstract method, :func:`~ModelInterface.DeviceFactory`,
    that has to be re-implemented by the driver modules.

    All methods of this class should return :class:`Reply` objects (except the
    :func:`DeviceFactory` function). However, for better readability, a decorator function
    :func:`to_reply` is provided, so that the types of the return values can be
    explicitly defined here.

    """

    # Class attributes
    version: str = "2.0"
    """Version of the :obj:`DriverInterface`."""

    idle_measurement_interval: Union[int, None] = None
    """The interval (in seconds) after which :func:`self.cmp_measure` will be executed, when idle."""

    # Instance attributes
    devmap: dict[tuple, "ModelDevice"]
    """Map of registered devices, the tuple keys are `component = (address, channel)`"""

    settings: dict[str, Any]
    """A settings map to contain driver-specific settings such as ``dllpath`` for BioLogic"""

    constants: dict[str, Any]
    """A map that should be populated with driver-specific run-time constants."""

    def __init__(self, settings=None):
        self.devmap = {}
        self.constants = {}
        self.settings = settings if settings is not None else {}

    @abstractmethod
    def DeviceFactory(self, key: Key, **kwargs) -> "ModelDevice":
        """
        A factory function which is used to pass this instance of the :class:`ModelInterface`
        to the new :class:`ModelDevice` instance.
        """
        pass

    def dev_register(self, **kwargs):
        logger.warning(
            "Use of 'dev_register' is deprecated in favour of 'cmp_register' "
            "and will stop working in tomato-3.0"
        )
        return self.cmp_register(**kwargs)

    def dev_teardown(self, **kwargs):
        logger.warning(
            "Use of 'dev_teardown' is deprecated in favour of 'cmp_teardown' "
            "and will stop working in tomato-3.0"
        )
        return self.cmp_teardown(**kwargs)

    def dev_reset(self, **kwargs):
        logger.warning(
            "Use of 'dev_reset' is deprecated in favour of 'cmp_reset' "
            "and will stop working in tomato-3.0"
        )
        return self.cmp_reset(**kwargs)

    def dev_set_attr(self, **kwargs):
        logger.warning(
            "Use of 'dev_set_attr' is deprecated in favour of 'cmp_set_attr' "
            "and will stop working in tomato-3.0"
        )
        return self.cmp_set_attr(**kwargs)

    def dev_get_attr(self, **kwargs):
        logger.warning(
            "Use of 'dev_get_attr' is deprecated in favour of 'cmp_get_attr' "
            "and will stop working in tomato-3.0"
        )
        return self.cmp_get_attr(**kwargs)

    def dev_status(self, **kwargs):
        logger.warning(
            "Use of 'dev_status' is deprecated in favour of 'cmp_status' "
            "and will stop working in tomato-3.0"
        )
        return self.cmp_status(**kwargs)

    def dev_capabilities(self, **kwargs):
        logger.warning(
            "Use of 'dev_capabilities' is deprecated in favour of 'cmp_capabilities' "
            "and will stop working in tomato-3.0"
        )
        return self.cmp_capabilities(**kwargs)

    def dev_attrs(self, **kwargs):
        logger.warning(
            "Use of 'dev_attrs' is deprecated in favour of 'cmp_attrs' "
            "and will stop working in tomato-3.0"
        )
        return self.cmp_attrs(**kwargs)

    @log_errors
    @to_reply
    def cmp_register(
        self, address: str, channel: str, **kwargs: dict
    ) -> tuple[bool, str, set]:
        """
        Register a new device component in this driver.

        Creates a :class:`ModelDevice` representing a device component, storing it in
        the :obj:`self.devmap` using the provided `address` and `channel`.

        Returns the :class:`set` of capabilities of the registered component as the
        :obj:`Reply.data`.
        """
        key = (address, channel)
        self.devmap[key] = self.DeviceFactory(key, **kwargs)
        capabs = self.devmap[key].capabilities()
        return (True, f"device {key!r} registered", capabs)

    @log_errors
    @to_reply
    @in_devmap
    def cmp_teardown(self, key: Key, **kwargs: dict) -> tuple[bool, str, None]:
        """
        Emergency stop function.

        Should set the device component into a documented, safe state. The function is
        to be only called in case of critical errors, or when the component is being
        removed, not as part of normal operation (i.e. it is not intended as a clean-up
        after task completion).
        """
        status = self.task_status(key=key, **kwargs)
        if status.data:
            logger.warning("tearing down component %s with a running task!", key)
            self.task_stop(key=key, **kwargs)
        self.dev_reset(key=key, **kwargs)
        del self.devmap[key]
        return (True, f"device {key!r} torn down", None)

    @log_errors
    @to_reply
    @in_devmap
    def cmp_reset(self, key: Key, **kwargs: dict) -> tuple[bool, str, None]:
        """
        Component reset function.

        Should set the device component into a documented, safe state. This function
        is executed at the end of every job.
        """
        self.devmap[key].reset()
        return (True, f"component {key!r} reset successfully", None)

    @log_errors
    @to_reply
    @in_devmap
    def cmp_set_attr(
        self, attr: str, val: Val, key: Key, **kwargs: dict
    ) -> tuple[bool, str, Val]:
        """
        Set value of the :class:`Attr` of the specified device component.

        Pass-through to the :func:`ModelDevice.set_attr` function. No type or
        read-write validation performed here! Returns the validated or coerced
        value as the :obj:`Reply.data`.
        """
        ret = self.devmap[key].set_attr(attr=attr, val=val, **kwargs)
        return (True, f"attr {attr!r} of component {key} set to {ret}", ret)

    @log_errors
    @to_reply
    @in_devmap
    def cmp_get_attr(
        self, attr: str, key: Key, **kwargs: dict
    ) -> tuple[bool, str, Val]:
        """
        Get value of the :class:`Attr` from the specified device component.

        Pass-through to the :func:`ModelDevice.get_attr` function. No type
        coercion is done here. Returns the value as the :obj:`Reply.data`.

        """
        ret = self.devmap[key].get_attr(attr=attr, **kwargs)
        return (True, f"attr {attr!r} of component {key} is: {ret}", ret)

    @log_errors
    @to_reply
    @in_devmap
    def cmp_status(self, key: Key, **kwargs: dict) -> tuple[bool, str, dict]:
        """
        Get the status report from the specified device component.

        Iterates over all :class:`Attrs` on the component that have ``status=True`` and
        returns their values in the :obj:`Reply.data` as a :class:`dict`.
        """
        ret = {}
        for k, attr in self.devmap[key].attrs(key=key, **kwargs).items():
            if attr.status:
                ret[k] = self.devmap[key].get_attr(attr=k, **kwargs)

        ret["running"] = self.devmap[key].running
        msg = f"component {key} is{' ' if ret['running'] else ' not '}running"
        return (True, msg, ret)

    @log_errors
    @to_reply
    @in_devmap
    def cmp_capabilities(self, key: Key, **kwargs) -> tuple[bool, str, set]:
        """
        Returns the capabilities of the device component.

        Pass-through to :func:`ModelDevice.capabilities`. Returns the :class:`set`
        of capabilities in :obj:`Reply.data`.
        """
        ret = self.devmap[key].capabilities(**kwargs)
        return (True, f"capabilities supported by component {key!r} are: {ret}", ret)

    @log_errors
    @to_reply
    @in_devmap
    def cmp_attrs(self, key: Key, **kwargs: dict) -> tuple[bool, str, dict]:
        """
        Query available :class:`Attrs` on the specified device component.

        Pass-through to the :func:`ModelDevice.attrs` function. Returns the
        :class:`dict` of attributes as the :obj:`Reply.data`.
        """
        ret = self.devmap[key].attrs(**kwargs)
        return (True, f"attrs of component {key!r} are: {ret}", ret)

    @log_errors
    @to_reply
    @in_devmap
    def cmp_constants(self, key: Key, **kwargs: dict) -> tuple[bool, str, dict]:
        """
        Query constants on the specified device component and this driver.

        Returns the :class:`dict` of constants as the :obj:`Reply.data`.
        """
        ret = self.constants | self.devmap[key].constants
        return (True, f"constants of component {key!r} are: {ret}", ret)

    @log_errors
    @to_reply
    @in_devmap
    def cmp_last_data(
        self, key: Key, **kwargs: dict
    ) -> tuple[bool, str, Union[None, xr.Dataset]]:
        """
        Fetch the last stored data on the component.

        Passthrough to :func:`ModelDevice.get_last_data`. The data in the form of
        a :class:`xarray.Dataset` is returned as the :obj:`Reply.data`.
        """
        ret = self.devmap[key].get_last_data(**kwargs)
        if ret is None:
            return (False, f"no data present on component {key!r}", None)
        else:
            return (True, f"last datapoint on component {key!r} at {ret.uts}", ret)

    @log_errors
    @to_reply
    @in_devmap
    def cmp_measure(self, key: Key, **kwargs: dict) -> tuple[bool, str, None]:
        """
        Do a single measurement on the component according to its current
        configuration.

        Fails if the component already has a running task / measurement.

        """
        if self.devmap[key].running:
            return (False, f"measurement already running on component {key!r}", None)
        else:
            self.devmap[key].measure()
            return (True, f"measurement started on component {key!r}", None)

    @log_errors
    @to_reply
    @in_devmap
    def task_start(
        self, key: Key, task: Task, **kwargs
    ) -> tuple[bool, str, Union[set, Task]]:
        """
        Submit a :class:`Task` onto the specified device component.

        Pushes the supplied :class:`Task` into the :class:`Queue` of the component,
        then starts the worker thread. Checks that the :class:`Task` is among the
        capabilities of this component.
        """
        ret = self.task_validate(key=key, task=task, **kwargs)
        if not ret.success:
            return ret

        logger.info("starting task '%s' on component %s", task.technique_name, key)
        self.devmap[key].task_list.put(task)
        self.devmap[key].run()
        logger.info("task '%s' on component %s started", task.technique_name, key)
        return (True, f"task {task!r} started successfully", task)

    @log_errors
    @to_reply
    @in_devmap
    def task_status(self, key: Key, **kwargs: dict) -> tuple[bool, str, dict]:
        """
        Returns the task readiness status of the specified device component.

        The `running` entry in the data slot of the :class:`Reply` indicates whether
        a :class:`Task` is running. The `can_submit` entry indicates whether another
        :class:`Task` can be queued onto the device component already.
        """
        running = self.devmap[key].running
        data = dict(running=running, can_submit=not running)
        if running:
            return (True, "running", data)
        else:
            return (True, "ready", data)

    @log_errors
    @to_reply
    @in_devmap
    def task_stop(
        self, key: Key, **kwargs
    ) -> tuple[bool, str, Union[xr.Dataset, None]]:
        """
        Stops a running task and returns any collected data.

        Pass-through to :func:`ModelDevice.stop_task` and :func:`ModelInterface.task_data`.

        If there is any cached data, it is returned as a :class:`xarray.Dataset` in the
        :obj:`Reply.data` and the cache is cleared.
        """
        ret = self.devmap[key].stop_task(**kwargs)
        if ret is not None:
            return (False, "failed to stop task", ret)
        else:
            ret = self.task_data(key=key)
            if ret.success:
                return (True, f"task stopped, {ret.msg}", ret.data)
            else:
                return (True, f"task stopped, {ret.msg}", None)

    @log_errors
    @to_reply
    @in_devmap
    def task_data(
        self, key: Key, **kwargs
    ) -> tuple[bool, str, Union[xr.Dataset, None]]:
        """
        Return cached task data on the device component and clean the cache.

        Pass-through for :func:`ModelDevice.get_data`, which should return a
        :class:`xarray.Dataset` that is fully annotated.

        This function gets called by the job thread every `device.pollrate`, it therefore
        incurs some IPC cost.

        """
        data = self.devmap[key].get_data(**kwargs)
        if data is None:
            return (False, "found no new datapoints", None)
        else:
            return (True, f"found {len(data)} new datapoints", data)

    @log_errors
    @to_reply
    @in_devmap
    def task_validate(self, key: Key, task: Task, **kwargs) -> tuple[bool, str, None]:
        """
        Validate the provided :class:`Task` for submission on the component
        identified by :obj:`key`.

        """
        logger.info("validating task '%s' on component %s", task.technique_name, key)
        if task.technique_name not in self.devmap[key].capabilities(**kwargs):
            msg = f"unknown task {task.technique_name!r} requested"
            return (False, msg, None)
        attrs = self.devmap[key].attrs(**kwargs)
        for par, val in task.task_params.items():
            if par not in attrs:
                msg = f"unknown attribute {par!r} cannot be set"
                return (False, msg, None)
            props = attrs[par]
            if not props.rw:
                msg = f"attribute {par!r} is read-only"
                return (False, msg, None)
            if not isinstance(val, props.type):
                try:
                    val = props.type(val)
                except (ValueError, pint.errors.UndefinedUnitError):
                    msg = f"could not coerce {par!r} to type {props.type}"
                    return (False, msg, None)
            if isinstance(val, pint.Quantity):
                if val.dimensionless and props.units is not None:
                    val = pint.Quantity(val.m, props.units)
                if val.dimensionality != pint.Quantity(props.units).dimensionality:
                    msg = f"attribute {par!r} has the wrong dimensionality {str(val.dimensionality)}"
                    return (False, msg, None)
            if props.minimum is not None and val < props.minimum:
                msg = f"attr {par!r} is smaller than {props.minimum}"
                return (False, msg, None)
            if props.maximum is not None and val > props.maximum:
                msg = f"attr {par!r} is greater than {props.maximum}"
                return (False, msg, None)
        return (True, "task validated successfully", None)

    @log_errors
    def status(self) -> Reply:
        """
        Returns the driver status. Currently that is the names of the components in
        the `devmap`.

        """
        devkeys = self.devmap.keys()
        return Reply(
            success=True,
            msg=f"driver running with {len(devkeys)} devices",
            data=dict(devkeys=devkeys),
        )

    @log_errors
    def reset(self) -> Reply:
        """
        Resets the driver.

        Called when the driver process is quitting. Instructs all remaining tasks to
        stop. Warns when devices linger. Passes through to :func:`dev_reset`. This is
        not a pass-through to :func:`dev_teardown`.

        """
        logger.info("resetting all components on this driver")
        for key, dev in self.devmap.items():
            if dev.thread.is_alive():
                logger.warning("stopping task on component %s", key)
                setattr(dev.thread, "do_run", False)
                dev.thread.join(timeout=1)
            if dev.thread.is_alive():
                logger.error("task on component %s is still running", key)
            else:
                logger.debug("component %s has no running task", key)
            self.dev_reset(key=key)
        return Reply(
            success=True,
            msg="all components on driver have been reset",
        )


class ModelDevice(metaclass=ABCMeta):
    """
    An abstract base class specifying a manager for an individual component.

    This class should handle determining attributes and capabilities of the component,
    the reading/writing of those attributes, processing of tasks, and caching and
    returning of task data.
    """

    driver: ModelInterface
    """The parent :class:`DriverInterface` instance."""

    constants: dict[str, Any]
    """Constant metadata of this component."""

    data: Union[xr.Dataset, None]
    """Container for cached data on this component."""

    last_data: Union[xr.Dataset, None]
    """Container for last datapoint on this component."""

    datalock: RLock
    """Lock object for thread-safe data manipulation."""

    key: Key
    """The key in :obj:`self.driver.devmap` referring to this object."""

    thread: Thread
    """The worker :class:`Thread`."""

    task_list: Queue
    """A :class:`Queue` used to pass :class:`Tasks` to the worker :class:`Thread`."""

    running: bool

    def __init__(self, driver, key, **kwargs) -> None:
        self.driver = driver
        self.key = key
        self.task_list = Queue()
        self.thread = Thread(target=self.task_runner, daemon=True)
        self.data = None
        self.last_data = None
        self.running = False
        self.datalock = RLock()
        self.constants = dict()
        atexit.register(self.reset)

    def run(self) -> None:
        """Helper function for starting the :obj:`self.thread` as a task."""
        self.thread.do_run = True
        self.thread.start()

    def measure(self) -> None:
        """Helper function for starting the :obj:`self.thread` as a measurement."""
        self.thread = Thread(target=self.meas_runner, daemon=True)
        self.thread.do_run = True
        self.thread.start()

    def task_runner(self) -> None:
        """
        Target function for the :obj:`self.thread` when handling :class:`Tasks`.

        This function waits for a :class:`Task` passed using :obj:`self.task_list`,
        then handles setting all :class:`Attrs` using the :func:`prepare_task`
        function, and finally handles the main loop of the task, periodically running
        the :func:`do_task` function (using `task.sampling_interval`) until the
        maximum task duration (i.e. `task.max_duration`) is exceeded.

        The :obj:`self.thread` is re-primed for future :class:`Tasks` at the end
        of this function.
        """
        e = None
        self.running = True
        thread = current_thread()
        try:
            task: Task = self.task_list.get()
            self.prepare_task(task=task)
        except Exception as e:
            logger.critical(e, exc_info=True)
            thread.do_run = False

        t_start = time.perf_counter()
        t_prev = t_start
        self.data = None
        while getattr(thread, "do_run"):
            t_now = time.perf_counter()
            if t_now - t_prev > task.sampling_interval:
                with self.datalock:
                    try:
                        self.do_task(task, t_start=t_start, t_now=t_now, t_prev=t_prev)
                    except Exception as e:
                        logger.critical(e, exc_info=True)
                        thread.do_run = False
                t_prev += task.sampling_interval
            if t_now - t_start > task.max_duration:
                break
            time.sleep(max(1e-2, task.sampling_interval / 20))

        self.task_list.task_done()
        self.running = False
        self.thread = Thread(target=self.task_runner, daemon=True)
        if e is None:
            logger.info(
                "task '%s' on component %s is done", task.technique_name, self.key
            )

    def meas_runner(self) -> None:
        """
        Target function for the :obj:`self.thread` when performing one shot measurements.

        Performs the measurement using :func:`self.do_measure`. Resets :obj:`self.thread`
        for future :class:`Tasks`.
        """
        try:
            self.running = True
            self.do_measure()
            self.running = False
            self.thread = Thread(target=self.task_runner, daemon=True)
            logger.debug("measurement on component %s is done", self.key)
        except Exception as e:
            logger.critical(e, exc_info=True)

    def prepare_task(self, task: Task, **kwargs: dict) -> None:
        """
        Given a :class:`Task`, prepare this component for execution by setting all
        :class:`Attrs` as specified in the `task.task_params` dictionary.
        """
        if task.task_params is not None:
            for k, v in task.task_params.items():
                self.set_attr(attr=k, val=v)

    def do_task(self, task: Task, **kwargs: dict) -> None:
        """
        Periodically called task execution function.

        This function is responsible for updating :obj:`self.data` with new data, i.e.
        performing the measurement. It should also update the value of :obj:`self.last_data`,
        so that the component status is consistent with the cached data.
        """
        self.do_measure(**kwargs)
        if self.data is None:
            self.data = self.last_data
        else:
            self.data = xr.concat(
                [self.data, self.last_data], dim="uts", data_vars="minimal"
            )

    @abstractmethod
    def do_measure(self, **kwargs: dict) -> None:
        """
        One shot execution worker function.

        This function is performs a measurement using the current configuration of
        :obj:`self.attrs`, and stores the result in :obj:`self.last_data`.
        """
        pass

    def stop_task(self, **kwargs: dict) -> None:
        """Stops the currently running task."""
        logger.info("stopping running task on component %s", self.key)
        setattr(self.thread, "do_run", False)

    @abstractmethod
    def set_attr(self, attr: str, val: Val, **kwargs: dict) -> Val:
        """
        Sets the specified :class:`Attr` to :obj:`val`.

        This function should handle any data type coercion and validation
        using e.g. :obj:`Attr.maximum` and :obj:`Attr.minimum`.

        Returns the coerced value corresponding to :obj:`val`.
        """
        pass

    @abstractmethod
    def get_attr(self, attr: str, **kwargs: dict) -> Val:
        """Reads the value of the specified :class:`Attr`."""
        pass

    def get_data(self, **kwargs: dict) -> xr.Dataset:
        """
        Returns the cached :obj:`self.data` as a :class:`xarray.Dataset` before
        clearing the cache.
        """
        with self.datalock:
            ret = self.data
            self.data = None
        return ret

    def get_last_data(self, **kwargs: dict) -> xr.Dataset:
        """Returns the :obj:`last_data` object as a :class:`xarray.Dataset`."""
        return self.last_data

    @abstractmethod
    def attrs(**kwargs) -> dict[str, Attr]:
        """Returns a :class:`dict` of all available :class:`Attrs`."""
        pass

    @abstractmethod
    def capabilities(**kwargs) -> set:
        """Returns a :class:`set` of all supported techniques."""
        pass

    def status(self, **kwargs) -> dict[str, Val]:
        """Compiles a status report from :class:`Attrs` marked as `status=True`."""
        status = {}
        for attr, props in self.attrs().items():
            if props.status:
                status[attr] = self.get_attr(attr)
        return status

    def reset(self, **kwargs) -> None:
        """Resets the component to an initial status."""
        logger.info("resetting component %s", self.key)
        self.task_list = Queue()
        self.thread = Thread(target=self.task_runner, daemon=True)
        self.data = None
        self.running = False
        self.datalock = RLock()



import zmq
import random
import subprocess

from . import utils
import tomato

PORT = 12345
TIME = 1000
CTXT = zmq.Context()
kwargs = dict(port=PORT, timeout=TIME, context=CTXT)


def test_passata_api_status(start_tomato_daemon, stop_tomato_daemon):
    utils.wait_until_tomato_running(port=PORT, timeout=1000)
    utils.wait_until_tomato_drivers(port=PORT, timeout=3000)
    ret = tomato.passata.status(
        name="example_counter:(example-addr,1)",
        **kwargs,
    )
    print(f"{ret=}")
    assert ret.success
    assert "running" in ret.data


def test_passata_api_attrs(start_tomato_daemon, stop_tomato_daemon):
    utils.wait_until_tomato_running(port=PORT, timeout=1000)
    utils.wait_until_tomato_drivers(port=PORT, timeout=3000)
    ret = tomato.passata.attrs(
        name="example_counter:(example-addr,1)",
        **kwargs,
    )
    print(f"{ret=}")
    assert ret.success
    assert "max" in ret.data


def test_passata_api_capabs(start_tomato_daemon, stop_tomato_daemon):
    utils.wait_until_tomato_running(port=PORT, timeout=1000)
    utils.wait_until_tomato_drivers(port=PORT, timeout=3000)
    ret = tomato.passata.capabilities(
        name="example_counter:(example-addr,1)",
        **kwargs,
    )
    print(f"{ret=}")
    assert ret.success
    assert "count" in ret.data


def test_passata_api_get_attrs(start_tomato_daemon, stop_tomato_daemon):
    utils.wait_until_tomato_running(port=PORT, timeout=1000)
    utils.wait_until_tomato_drivers(port=PORT, timeout=3000)
    ret = tomato.passata.get_attrs(
        name="example_counter:(example-addr,1)",
        attrs=["max", "min"],
        **kwargs,
    )
    print(f"{ret=}")
    assert ret.success
    assert "max" in ret.data
    assert "min" in ret.data


def test_passata_api_set_attr(start_tomato_daemon, stop_tomato_daemon):
    utils.wait_until_tomato_running(port=PORT, timeout=1000)
    utils.wait_until_tomato_drivers(port=PORT, timeout=3000)
    val = random.random() * 100
    ret = tomato.passata.set_attr(
        name="example_counter:(example-addr,1)",
        attr="max",
        val=val,
        **kwargs,
    )
    print(f"{ret=}")
    assert ret.success
    assert ret.data == val
    ret = tomato.passata.get_attrs(
        name="example_counter:(example-addr,1)",
        attrs=["max"],
        **kwargs,
    )
    print(f"{ret=}")
    assert ret.success
    assert ret.data["max"] == val


def test_passata_api_reset(start_tomato_daemon, stop_tomato_daemon):
    utils.wait_until_tomato_running(port=PORT, timeout=1000)
    utils.wait_until_tomato_drivers(port=PORT, timeout=3000)
    ret = tomato.passata.reset(
        name="example_counter:(example-addr,1)",
        **kwargs,
    )
    print(f"{ret=}")
    assert ret.success


def test_passata_api_constants(start_tomato_daemon, stop_tomato_daemon):
    utils.wait_until_tomato_running(port=PORT, timeout=1000)
    utils.wait_until_tomato_drivers(port=PORT, timeout=3000)
    ret = tomato.passata.constants(
        name="example_counter:(example-addr,1)",
        **kwargs,
    )
    print(f"{ret=}")
    assert ret.success
    assert ret.data["example_meta"] == "example string"


def test_passata_api_last_data(start_tomato_daemon, stop_tomato_daemon):
    utils.wait_until_tomato_running(port=PORT, timeout=1000)
    utils.wait_until_tomato_drivers(port=PORT, timeout=3000)
    ret = tomato.passata.get_last_data(
        name="example_counter:(example-addr,1)",
        **kwargs,
    )
    print(f"{ret=}")
    assert ret.success
    assert "uts" in ret.data.coords


def test_passata_api_measure(start_tomato_daemon, stop_tomato_daemon):
    utils.wait_until_tomato_running(port=PORT, timeout=1000)
    utils.wait_until_tomato_drivers(port=PORT, timeout=3000)
    ret = tomato.passata.measure(
        name="example_counter:(example-addr,1)",
        **kwargs,
    )
    assert ret.success


def test_passata_api_force(datadir, start_tomato_daemon, stop_tomato_daemon):
    utils.wait_until_tomato_running(port=PORT, timeout=1000)
    utils.wait_until_tomato_drivers(port=PORT, timeout=3000)
    utils.run_casenames(["counter_5_0.2"], [None], ["pip-counter"])
    assert utils.wait_until_ketchup_status(jobid=1, status="r", port=PORT, timeout=5000)

    ret = tomato.passata.set_attr(
        name="example_counter:(example-addr,1)",
        attr="max",
        val=15,
        force=False,
        **kwargs,
    )
    assert ret.success is False
    assert "running component" in ret.msg

    ret = tomato.passata.set_attr(
        name="example_counter:(example-addr,1)",
        attr="max",
        val=15,
        force=True,
        **kwargs,
    )
    assert ret.success
    assert "set to 15.0" in ret.msg


def test_passata_cli(start_tomato_daemon, stop_tomato_daemon):
    utils.wait_until_tomato_running(port=PORT, timeout=1000)
    utils.wait_until_tomato_drivers(port=PORT, timeout=3000)
    ret = subprocess.run(
        [
            "passata",
            "status",
            "example_counter:(example-addr,1)",
            "-p",
            f"{PORT}",
        ],
        capture_output=True,
        text=True,
    )
    print(f"{ret=}")
    assert "Success: component ('example-addr', '1')" in ret.stdout

    ret = subprocess.run(
        [
            "passata",
            "attrs",
            "example_counter:(example-addr,1)",
            "-p",
            f"{PORT}",
        ],
        capture_output=True,
        text=True,
    )
    print(f"{ret=}")
    assert "Success: attrs of component ('example-addr', '1') are" in ret.stdout

    ret = subprocess.run(
        [
            "passata",
            "capabilities",
            "example_counter:(example-addr,1)",
            "-p",
            f"{PORT}",
        ],
        capture_output=True,
        text=True,
    )
    print(f"{ret=}")
    assert (
        "Success: capabilities supported by component ('example-addr', '1') are"
        in ret.stdout
    )

    ret = subprocess.run(
        [
            "passata",
            "get",
            "example_counter:(example-addr,1)",
            "max",
            "-p",
            f"{PORT}",
        ],
        capture_output=True,
        text=True,
    )
    print(f"{ret=}")
    assert (
        "Success: attr 'max' of component 'example_counter:(example-addr,1)' is"
        in ret.stdout
    )

    ret = subprocess.run(
        [
            "passata",
            "constants",
            "example_counter:(example-addr,1)",
            "max",
            "-p",
            f"{PORT}",
        ],
        capture_output=True,
        text=True,
    )
    print(f"{ret=}")
    assert "Success: constants of component ('example-addr', '1') are" in ret.stdout



import zmq
from tomato import tomato
from tomato.models import Reply, Component, Driver
from typing import Any


def _name_to_cmp(
    name: str,
    port: int,
    timeout: int,
    context: zmq.Context,
) -> tuple[Component, Driver]:
    ret = tomato.status(
        port=port, timeout=timeout, context=context, stgrp="tomato", yaml=True
    )
    if not ret.success:
        return ret
    if name not in ret.data.cmps:
        return Reply(
            success=False,
            msg=f"component {name!r} not found on tomato",
            data=ret.data,
        )
    cmp = ret.data.cmps[name]
    drv = ret.data.drvs[cmp.driver]
    return cmp, drv


def _running_or_force(
    name: str,
    port: int,
    timeout: int,
    context: zmq.Context,
    force: bool,
) -> Reply:
    if not force:
        ret = status(port=port, timeout=timeout, context=context, name=name).data
        if ret["running"]:
            return Reply(
                success=False,
                msg=f"will not 'set_attr' on a running component {name!r}",
                data=None,
            )
    return Reply(success=True)


def status(
    *,
    port: int,
    timeout: int,
    context: zmq.Context,
    name: str,
    **_: dict,
) -> Reply:
    ret = _name_to_cmp(name, port, timeout, context)
    if isinstance(ret, Reply):
        return ret
    cmp, drv = ret

    kwargs = dict(channel=cmp.channel, address=cmp.address)
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{drv.port}")
    if drv.version == "1.0":
        req.send_pyobj(dict(cmd="dev_status", params={**kwargs}))
    else:
        req.send_pyobj(dict(cmd="cmp_status", params={**kwargs}))
    ret = req.recv_pyobj()
    req.close()
    return ret


def register(
    *,
    port: int,
    timeout: int,
    context: zmq.Context,
    name: str,
    **_: dict,
) -> Reply:
    ret = _name_to_cmp(name, port, timeout, context)
    if isinstance(ret, Reply):
        return ret
    cmp, drv = ret

    kwargs = dict(channel=cmp.channel, address=cmp.address)
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{drv.port}")
    if drv.version == "1.0":
        req.send_pyobj(dict(cmd="dev_register", params={**kwargs}))
    else:
        req.send_pyobj(dict(cmd="cmp_register", params={**kwargs}))
    ret = req.recv_pyobj()
    req.close()
    return ret


def attrs(
    *,
    port: int,
    timeout: int,
    context: zmq.Context,
    name: str,
    **_: dict,
) -> Reply:
    ret = _name_to_cmp(name, port, timeout, context)
    if isinstance(ret, Reply):
        return ret
    cmp, drv = ret

    kwargs = dict(channel=cmp.channel, address=cmp.address)
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{drv.port}")
    if drv.version == "1.0":
        req.send_pyobj(dict(cmd="attrs", params={**kwargs}))
    else:
        req.send_pyobj(dict(cmd="cmp_attrs", params={**kwargs}))
    ret = req.recv_pyobj()
    return ret


def capabilities(
    *,
    port: int,
    timeout: int,
    context: zmq.Context,
    name: str,
    **_: dict,
) -> Reply:
    ret = _name_to_cmp(name, port, timeout, context)
    if isinstance(ret, Reply):
        return ret
    cmp, drv = ret

    kwargs = dict(channel=cmp.channel, address=cmp.address)
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{drv.port}")
    if drv.version == "1.0":
        req.send_pyobj(dict(cmd="capabilities", params={**kwargs}))
    else:
        req.send_pyobj(dict(cmd="cmp_capabilities", params={**kwargs}))
    ret = req.recv_pyobj()
    return ret


def constants(
    *,
    port: int,
    timeout: int,
    context: zmq.Context,
    name: str,
    **_: dict,
) -> Reply:
    ret = _name_to_cmp(name, port, timeout, context)
    if isinstance(ret, Reply):
        return ret
    cmp, drv = ret

    if drv.version == "1.0":
        return Reply(
            success=False,
            msg=f"driver of component {name!r} is on version {drv.version}",
            data=None,
        )

    kwargs = dict(channel=cmp.channel, address=cmp.address)
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{drv.port}")
    req.send_pyobj(dict(cmd="cmp_constants", params={**kwargs}))
    ret = req.recv_pyobj()
    return ret


def get_attrs(
    port: int,
    timeout: int,
    context: zmq.Context,
    name: str,
    attrs: list[str],
    yaml: bool = False,
    **_: dict,
) -> Reply:
    ret = _name_to_cmp(name, port, timeout, context)
    if isinstance(ret, Reply):
        return ret
    cmp, drv = ret

    kwargs = dict(channel=cmp.channel, address=cmp.address)
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{drv.port}")
    data = dict()
    msg = ""
    for attr in attrs:
        if drv.version == "1.0":
            req.send_pyobj(dict(cmd="dev_get_attr", params={"attr": attr, **kwargs}))
        else:
            req.send_pyobj(dict(cmd="cmp_get_attr", params={"attr": attr, **kwargs}))
        ret = req.recv_pyobj()
        if not ret.success:
            return ret
        data[attr] = ret.data
        msg += f"attr {attr!r} of component {name!r} is: {ret.data}\n         "
    if yaml:
        msg = f"attrs {list(data.keys())} of component {name!r} retrieved"
    else:
        msg = msg.rstrip()
    return Reply(
        success=True,
        msg=msg,
        data=data,
    )


def set_attr(
    port: int,
    timeout: int,
    context: zmq.Context,
    name: str,
    attr: str,
    val: Any,
    force: bool = False,
    **_: dict,
) -> Reply:
    ret = _name_to_cmp(name, port, timeout, context)
    if isinstance(ret, Reply):
        return ret
    cmp, drv = ret
    ret = _running_or_force(name, port, timeout, context, force)
    if not ret.success:
        return ret

    kwargs = dict(channel=cmp.channel, address=cmp.address)
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{drv.port}")
    if drv.version == "1.0":
        req.send_pyobj(
            dict(cmd="dev_set_attr", params={"attr": attr, "val": val, **kwargs})
        )
    else:
        req.send_pyobj(
            dict(cmd="cmp_set_attr", params={"attr": attr, "val": val, **kwargs})
        )
    ret = req.recv_pyobj()
    return ret


def reset(
    *,
    port: int,
    timeout: int,
    context: zmq.Context,
    name: str,
    force: bool = False,
    **_: dict,
) -> Reply:
    ret = _name_to_cmp(name, port, timeout, context)
    if isinstance(ret, Reply):
        return ret
    cmp, drv = ret
    ret = _running_or_force(name, port, timeout, context, force)
    if not ret.success:
        return ret

    kwargs = dict(channel=cmp.channel, address=cmp.address)
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{drv.port}")
    if drv.version == "1.0":
        req.send_pyobj(dict(cmd="dev_reset", params=kwargs))
    else:
        req.send_pyobj(dict(cmd="cmp_reset", params=kwargs))
    ret = req.recv_pyobj()
    return ret


def get_last_data(
    *,
    port: int,
    timeout: int,
    context: zmq.Context,
    name: str,
    **_: dict,
) -> Reply:
    ret = _name_to_cmp(name, port, timeout, context)
    if isinstance(ret, Reply):
        return ret
    cmp, drv = ret

    if drv.version == "1.0":
        return Reply(
            success=False,
            msg=f"driver of component {name!r} is on version {drv.version}",
            data=None,
        )

    kwargs = dict(channel=cmp.channel, address=cmp.address)
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{drv.port}")
    req.send_pyobj(dict(cmd="cmp_last_data", params=kwargs))
    ret = req.recv_pyobj()
    return ret


def measure(
    *,
    port: int,
    timeout: int,
    context: zmq.Context,
    name: str,
    **_: dict,
) -> Reply:
    ret = _name_to_cmp(name, port, timeout, context)
    if isinstance(ret, Reply):
        return ret
    cmp, drv = ret

    if drv.version == "1.0":
        return Reply(
            success=False,
            msg=f"driver of component {name!r} is on version {drv.version}",
            data=None,
        )

    kwargs = dict(channel=cmp.channel, address=cmp.address)
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{drv.port}")
    req.send_pyobj(dict(cmd="cmp_measure", params=kwargs))
    ret = req.recv_pyobj()
    return ret



"""
**tomato.daemon.job**: the job manager of tomato daemon
-------------------------------------------------------
.. codeauthor::
    Peter Kraus

.. note::

    Functions in this module that receive the :class:`~tomato.models.Daemon` state
    object should be acting on a copy. All changes to the :class:`Daemon` state have to
    be propagated via the :class:`tomato.daemon.cmd` set of functions.

"""

import os
import subprocess
import logging
import json
import time
import argparse
from importlib import metadata
from datetime import datetime, timezone
from pathlib import Path
from threading import current_thread, Thread
import zmq
import psutil
import sys

from tomato.daemon.io import merge_netcdfs, data_to_pickle
from tomato.daemon import jobdb
from tomato.models import Pipeline, Daemon, Component, Device, Driver, Job
from dgbowl_schemas.tomato import to_payload
from dgbowl_schemas.tomato.payload import Task

logger = logging.getLogger(__name__)


def method_validate(
    method: list[Task],
    pip: Pipeline,
    drvs: dict[str, Driver],
    cmps: dict[str, Component],
    context: zmq.Context,
):
    for task in method:
        for cmp in pip.components:
            if task.technique_name in cmps[cmp].capabilities:
                drv = drvs[cmps[cmp].driver]
                if drv.version == "1.0":
                    logger.info("cannot validate task using DriverInterface-1.0")
                    break
                req = context.socket(zmq.REQ)
                req.connect(f"tcp://127.0.0.1:{drv.port}")
                params = dict(
                    task=task,
                    address=cmps[cmp].address,
                    channel=cmps[cmp].channel,
                )
                req.send_pyobj(dict(cmd="task_validate", params=params))
                ret = req.recv_pyobj()
                req.close()
                if ret.success:
                    break
        else:
            return False
    return True


def find_matching_pipelines(
    pips: dict[str, Pipeline],
    cmps: dict[str, Component],
    drvs: dict[str, Driver],
    method: list[Task],
    context: zmq.Context,
) -> list[Pipeline]:
    req_roles = set([item.component_role for item in method])
    req_capabs = set([item.technique_name for item in method])

    candidates = []
    for pip in pips.values():
        roles = set()
        capabs = set()
        for comp in pip.components:
            c = cmps[comp]
            roles.add(c.role)
            capabs.update(c.capabilities)
        if req_roles.intersection(roles) == req_roles:
            if req_capabs.intersection(capabs) == req_capabs:
                if method_validate(method, pip, drvs, cmps, context):
                    candidates.append(pip)
    return candidates


def kill_tomato_job(process: psutil.Process):
    """
    Wrapper around :func:`psutil.terminate`.

    Here we kill the (grand)children of the process with the name of `tomato-job`,
    i.e. the individual task functions. This allows the `tomato-job` process to exit
    gracefully once the task functions join.

    Note that on Windows, the `tomato-job.exe` process has two children: a `python.exe`
    which is the actual process running the job, and `conhost.exe`, which we want to
    avoid killing.

    """
    logger = logging.getLogger(f"{__name__}.kill_tomato_job")
    if psutil.WINDOWS:
        pc = [p for p in process.children() if p.name() not in {"conhost.exe"}]
        to_kill = []
        for child in pc:
            to_kill += child.children()
    elif psutil.POSIX:
        to_kill = [p for p in process.children()]
    for proc in to_kill:
        logger.warning(f"killing process {proc.name()!r} with pid {proc.pid}")
        proc.terminate()
    gone, alive = psutil.wait_procs(to_kill, timeout=1)
    logger.debug(f"{gone=}")
    logger.debug(f"{alive=}")


def manage_running_pips(pips: dict, dbpath: str, req):
    """
    Function that manages jobs and `tomato-daemon` pipelines.

    The function only affects pipelines marked as running, i.e. with a set ``jobid``.
    Jobs scheduled for killing (i.e. ``status == "rd"``) are terminated. Jobs that
    are supposed to be running but have crashed are given appropriate status (``"ce"``).
    Pipelines of both are reset.

    Successful job completions are not processed here, but within the job process.

    """
    logger = logging.getLogger(f"{__name__}.manage_running_pips")
    running = [pip for pip in pips.values() if pip.jobid is not None]
    logger.debug(f"{running=}")
    for pip in running:
        job = jobdb.get_job_id(pip.jobid, dbpath)
        if job.pid is None:
            continue
        pidexists = psutil.pid_exists(job.pid)
        if pidexists:
            pidexists = psutil.Process(job.pid).status() is not psutil.STATUS_ZOMBIE
        logger.debug(f"{pidexists=}")
        reset = False
        # running jobs scheduled for killing (status == 'rd') should be killed
        if pidexists and job.status == "rd":
            logger.info(f"job {job.id} with pid {job.pid} will be terminated")
            proc = psutil.Process(pid=job.pid)
            kill_tomato_job(proc)
            logger.info(f"job {job.id} with pid {job.pid} was terminated successfully")
            merge_netcdfs(job)
            reset = True
            params = dict(status="cd")
        # dead jobs marked as running (status == 'r') should be cleared
        elif (not pidexists) and job.status == "r":
            logger.warning(f"the pid {job.pid} of job {job.id} has not been found")
            reset = True
            params = dict(status="ce")
        # pipelines of completed jobs should be reset
        elif (not pidexists) and job.status == "c":
            logger.info(f"the pid {job.pid} of job {job.id} has not been found")
            ready = job.payload.settings.unlock_when_done
            params = dict(jobid=None, ready=ready, name=pip.name)
            req.send_pyobj(dict(cmd="pipeline", params=params))
            ret = req.recv_pyobj()
            logger.debug(f"{ret=}")
        if reset:
            params["pid"] = None
            params["completed_at"] = str(datetime.now(timezone.utc))
            jobdb.update_job_id(job.id, params, dbpath)
            logger.debug(f"pipeline {pip.name!r} will be reset")
            params = dict(jobid=None, ready=False, name=pip.name)
            req.send_pyobj(dict(cmd="pipeline", params=params))
            ret = req.recv_pyobj()
            if not ret.success:
                logger.error(f"could not set params {params} on pip: {pip.name!r}")
                continue


def check_queued_jobs(
    pips: dict,
    cmps: dict,
    drvs: dict,
    dbpath: str,
    context: zmq.Context,
) -> dict[int, list[Pipeline]]:
    """
    Function to check whether the queued jobs can be submitted onto any pipeline.

    Returns a :class:`dict` containing the jobids as keys and lists of matched
    :class:`Pipelines` as values.
    """
    logger = logging.getLogger(f"{__name__}.check_queued_jobs")
    matched = {}
    queue = jobdb.get_jobs_where("status IN ('q', 'qw')", dbpath)
    for job in queue:
        matched[job.id] = find_matching_pipelines(
            pips, cmps, drvs, job.payload.method, context
        )
        if len(matched[job.id]) > 0 and job.status == "q":
            logger.info(
                "job %d can queue on pips: {%s}",
                job.id,
                [p.name for p in matched[job.id]],
            )
            params = dict(status="qw")
            job = jobdb.update_job_id(job.id, params, dbpath)
    return matched


def action_queued_jobs(daemon, matched, req):
    """
    Function that assigns jobs if a matched pipeline contains the requested sample.

    The `tomato-job` process is launched from this function.
    """
    logger = logging.getLogger(f"{__name__}.action_queued_jobs")
    for jobid in sorted(matched.keys()):
        job = jobdb.get_job_id(jobid, daemon.settings["jobs"]["dbpath"])
        for pip in matched[job.id]:
            if not pip.ready:
                continue
            elif pip.sampleid != job.payload.sample.name:
                continue
            logger.info(f"job {job.id} found a matched & ready pip: {pip.name!r}")
            params = dict(jobid=job.id, ready=False, name=pip.name)
            req.send_pyobj(dict(cmd="pipeline", params=params))
            ret = req.recv_pyobj()
            if not ret.success:
                logger.error(f"could not set params {params} on pip: {pip.name!r}")
                continue
            else:
                pip.ready = False

            root = Path(daemon.settings["jobs"]["storage"]) / str(job.id)
            os.makedirs(root)

            jpath = root / "jobdata.json"
            jobargs = {
                "pipeline": pip.model_dump(),
                "payload": job.payload.model_dump(),
                "devices": {dn: dev.model_dump() for dn, dev in daemon.devs.items()},
                "job": dict(id=job.id, path=str(root)),
            }

            with jpath.open("w", encoding="UTF-8") as of:
                json.dump(jobargs, of, indent=1)

            cmd = [
                "tomato-job",
                "--port",
                str(daemon.port),
                "--verbosity",
                str(daemon.verbosity),
                "--dbpath",
                str(daemon.settings["jobs"]["dbpath"]),
                str(jpath),
            ]
            if psutil.WINDOWS:
                cfs = subprocess.CREATE_NO_WINDOW
                cfs |= subprocess.CREATE_NEW_PROCESS_GROUP
                subprocess.Popen(cmd, creationflags=cfs)
            elif psutil.POSIX:
                subprocess.Popen(cmd, start_new_session=True)
            logger.info(f"job {jobid} started on pip: {pip.name!r} and path: {jpath!r}")
            break


def manager(port: int, timeout: int = 500):
    """
    The job manager thread of `tomato-daemon`.

    This manager ensures the job queue is iterated over and pipelines are managed/reset.
    Note that we poll the `tomato-daemon` for status only once per iteration of the main
    loop.
    """
    context = zmq.Context()
    logger = logging.getLogger(f"{__name__}.manager")
    thread = current_thread()
    logger.info("launched successfully")
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{port}")
    poller = zmq.Poller()
    poller.register(req, zmq.POLLIN)
    to = timeout
    while getattr(thread, "do_run"):
        logger.debug("tick")
        req.send_pyobj(dict(cmd="status", sender=f"{__name__}.manager"))
        events = dict(poller.poll(to))
        if req not in events:
            logger.warning(f"could not contact tomato-daemon in {to} ms")
            to = to * 2
            continue
        elif to > timeout:
            to = timeout
        daemon = req.recv_pyobj().data
        dbpath = daemon.settings["jobs"]["dbpath"]
        manage_running_pips(daemon.pips, dbpath, req)
        matched_pips = check_queued_jobs(
            daemon.pips, daemon.cmps, daemon.drvs, dbpath, context
        )
        action_queued_jobs(daemon, matched_pips, req)
        time.sleep(timeout / 1e3)
    logger.info("instructed to quit")


def tomato_job() -> None:
    """
    The function called when `tomato-job` is executed.

    This function is resposible for managing all activities of a single job, including
    contacting the daemon about job pid, spawning of sub-processes to run tasks on each
    Component of the Pipeline, merging data at the end of the job, and reporting back
    to the daemon once the job is successfully finished.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s version {metadata.version('tomato')}",
    )
    parser.add_argument(
        "--port",
        help="Port on which tomato-daemon is listening.",
        default=1234,
        type=int,
    )
    parser.add_argument(
        "--timeout",
        help="Timeout [ms] for driver actions.",
        default=1000,
        type=int,
    )
    parser.add_argument(
        "--retries",
        help="Number of retries for driver actions.",
        default=10,
        type=int,
    )
    parser.add_argument(
        "--verbosity",
        help="Verbosity of the tomato-job.",
        default=logging.INFO,
        type=int,
    )
    parser.add_argument(
        "--dbpath",
        help="Path to the sqlite3 job database.",
        type=str,
    )
    parser.add_argument(
        "jobfile",
        type=Path,
        help="Path to a ketchup-processed payload json file.",
    )
    args = parser.parse_args()

    with args.jobfile.open() as infile:
        jsdata = json.load(infile)
    payload = to_payload(**jsdata["payload"])

    pip = jsdata["pipeline"]["name"]
    jobid = jsdata["job"]["id"]
    jobpath = Path(jsdata["job"]["path"]).resolve()

    logpath = jobpath / f"job-{jobid}.log"
    logging.basicConfig(
        level=args.verbosity,
        format="%(asctime)s - %(levelname)8s - %(name)-30s - %(message)s",
        handlers=[logging.FileHandler(logpath, mode="a")],
    )
    logger = logging.getLogger(__name__)

    logger.debug(f"{payload=}")

    verbosity = payload.settings.verbosity
    loglevel = logging._checkLevel(verbosity)
    logger.debug("setting logger verbosity to '%s'", verbosity)
    logger.setLevel(loglevel)

    if psutil.WINDOWS:
        pid = os.getppid()
    elif psutil.POSIX:
        pid = os.getpid()

    logger.info(f"assigning job {jobid} with pid {pid} into pipeline {pip!r}")
    context = zmq.Context()

    params = dict(pid=pid, status="r", executed_at=str(datetime.now(timezone.utc)))
    job = jobdb.update_job_id(jobid, params, args.dbpath)

    output = payload.settings.output
    outpath = Path(output.path)
    logger.info(f"output folder is {outpath}")
    if outpath.exists():
        assert outpath.is_dir()
    else:
        logger.debug("path does not exist, creating")
        os.makedirs(outpath)
    prefix = f"results.{jobid}" if output.prefix is None else output.prefix
    respath = outpath / f"{prefix}.nc"
    snappath = outpath / f"snapshot.{jobid}.nc"
    params = dict(respath=str(respath), snappath=str(snappath), jobpath=str(jobpath))
    job = jobdb.update_job_id(jobid, params, args.dbpath)

    logger.info("handing off to 'job_main_loop'")
    logger.info("==============================")
    ret = job_main_loop(context, args.port, job, pip, logpath)
    logger.info("==============================")

    job.completed_at = str(datetime.now(timezone.utc))

    if ret is None:
        job.status = "c"
    else:
        job.status = "ce"
    logger.info("writing final data to a NetCDF file")
    merge_netcdfs(job)
    logger.info("job finished with status '%s', updating job db", job.status)
    params = dict(status=job.status, completed_at=job.completed_at)
    job = jobdb.update_job_id(job.id, params, args.dbpath)
    logger.debug(f"{job=}")
    logger.info("exiting tomato-job")


def job_thread(
    tasks: list,
    component: Component,
    device: Device,
    driver: Driver,
    jobpath: Path,
    logpath: Path,
):
    """
    A subthread of `tomato-job`, responsible for tasks on one Component of a Pipeline.

    For each task in tasks, starts the task, then monitors the Component status and polls
    for data, and moves on to the next task as instructed in the payload.

    Stores the data for that Component as a `pickle` of a :class:`xr.Dataset`.
    """
    thread = current_thread()
    sender = f"{__name__}.job_thread({thread.ident})"
    logger = logging.getLogger(sender)
    logger.debug(f"in job thread of {component.role!r}")

    context = zmq.Context()
    req = context.socket(zmq.REQ)
    req.RCVTIMEO = 1000
    req.connect(f"tcp://127.0.0.1:{driver.port}")
    logger.info(f"job thread of {component.role!r} connected to tomato-daemon")

    kwargs = dict(address=component.address, channel=component.channel)

    datapath = Path(jobpath) / f"{component.role}.pkl"
    logger.debug("distributing tasks:")
    for ti, task in enumerate(tasks):
        thread.current_task = task
        logger.info("processing task %s:%d", component.role, ti)
        while True:
            time.sleep(1e-1)
            if task.start_with_task_name is None:
                pass
            elif task.start_with_task_name in thread.started_task_names:
                pass
            else:
                logger.debug("waiting for task_name '%s'", task.start_with_task_name)
                continue
            logger.debug("polling component '%s' for task readiness", component.role)
            req.send_pyobj(dict(cmd="task_status", params={**kwargs}))
            ret = req.recv_pyobj()
            if ret.success and ret.data["can_submit"]:
                break
            logger.warning("cannot submit onto component '%s', waiting", component.role)
        logger.info("sending task %s:%d to component", component.role, ti)
        req.send_pyobj(dict(cmd="task_start", params={"task": task, **kwargs}))
        ret = req.recv_pyobj()

        t0 = time.perf_counter()
        while True:
            tN = time.perf_counter()
            if tN - t0 > device.pollrate:
                logger.debug("polling task %s:%d for data", component.role, ti)
                try:
                    req.send_pyobj(dict(cmd="task_data", params={**kwargs}))
                    ret = req.recv_pyobj()
                except zmq.ZMQError as e:
                    logger.critical(e, exc_info=True)
                    thread.crashed = True
                    sys.exit(e)
                if ret.success:
                    logger.debug("pickling received data")
                    ds = ret.data
                    ds.attrs["tomato_Component"] = component.model_dump_json()
                    data_to_pickle(ds, datapath, role=component.role)
                t0 += device.pollrate

            logger.debug("polling task %s:%d for completion", component.role, ti)
            try:
                req.send_pyobj(dict(cmd="task_status", params={**kwargs}))
                ret = req.recv_pyobj()
            except zmq.ZMQError as e:
                logger.critical(e, exc_info=True)
                thread.crashed = True
                sys.exit(e)
            if ret.success and not ret.data["running"]:
                logger.info("task %s:%d no longer running, break", component.role, ti)
                break
            elif ret.success is False:
                logger.critical(f"{ret=}")
                break

            if (
                task.stop_with_task_name is not None
                and task.stop_with_task_name in thread.started_task_names
            ):
                logger.info("task %s:%d stop trigger met", component.role, ti)
                req.send_pyobj(dict(cmd="task_stop", params={**kwargs}))
                ret = req.recv_pyobj()
                logger.debug(f"{ret=}")
                break

            time.sleep(max(1e-1, (device.pollrate - (tN - t0)) / 2))
        logger.info("task %s:%d fetching final data", component.role, ti)
        req.send_pyobj(dict(cmd="task_data", params={**kwargs}))
        ret = req.recv_pyobj()
        if ret.success:
            data_to_pickle(ret.data, datapath, role=component.role)
        thread.completed_tasks.append(task)
        thread.current_task = None
    logger.info("all tasks done on component '%s', resetting", component.role)
    if driver.version == "1.0":
        req.send_pyobj(dict(cmd="dev_reset", params={**kwargs}))
    else:
        req.send_pyobj(dict(cmd="cmp_reset", params={**kwargs}))
    ret = req.recv_pyobj()
    if not ret.success:
        logger.warning("could not reset component '%s': %s", component.role, ret.msg)


def job_main_loop(
    context: zmq.Context,
    port: int,
    job: Job,
    pipname: str,
    logpath: Path,
) -> None:
    """
    The main loop function of `tomato-job`, split for better readability.
    """
    sender = f"{__name__}.job_main_loop"
    logger = logging.getLogger(sender)
    logger.debug("process started")

    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{port}")

    while True:
        req.send_pyobj(dict(cmd="status", sender=sender))
        daemon: Daemon = req.recv_pyobj().data
        if all([drv.port is not None for drv in daemon.drvs.values()]):
            break
        else:
            logger.debug("not all tomato-drivers have a port, waiting")
            time.sleep(1)

    pipeline = daemon.pips[pipname]
    logger.debug(f"{pipeline=}")
    logger.debug(f"{job=}")

    # collate steps by role
    plan = {}
    for step in job.payload.method:
        if step.component_role not in plan:
            plan[step.component_role] = []
        plan[step.component_role].append(step)
    logger.debug(f"{plan=}")

    # distribute plan into threads
    threads = {}
    for cmpk in pipeline.components:
        component = daemon.cmps[cmpk]
        logger.debug(f"{component=}")
        if component.role not in plan:
            continue
        tasks = plan[component.role]
        logger.debug(" tasks=%s", tasks)
        device = daemon.devs[component.device]
        logger.debug(" device=%s", device)
        driver = daemon.drvs[component.driver]
        logger.debug(" driver=%s", driver)
        threads[component.role] = Thread(
            target=job_thread,
            args=(tasks, component, device, driver, job.jobpath, logpath),
            name="job-thread",
        )
        threads[component.role].crashed = False
        threads[component.role].completed_tasks = []
        threads[component.role].current_task = None
        threads[component.role].started_task_names = set()
        threads[component.role].start()

    # wait until threads join or we're killed
    snapshot = job.payload.settings.snapshot
    t0 = time.perf_counter()
    started_task_names = set()
    while True:
        tN = time.perf_counter()
        if snapshot is not None and tN - t0 > snapshot.frequency:
            logger.debug("creating snapshot")
            merge_netcdfs(job, snapshot=True)
            t0 += snapshot.frequency

        # Collect and push task names
        for t in threads.values():
            if t.current_task is not None and t.current_task.task_name is not None:
                started_task_names.add(t.current_task.task_name)
        logger.debug("started task names are: %s", started_task_names)
        for t in threads.values():
            t.started_task_names.update(started_task_names)
        crashed = [t.crashed for t in threads.values()]
        joined = [t.is_alive() is False or t.crashed for t in threads.values()]
        if all(joined):
            if any(crashed):
                return 1
            else:
                return None
        else:
            # We'd like to execute this loop exactly once every second
            time.sleep(1.0 - tN % 1)



"""
**tomato.drivers**: Shim interfacing with tomato driver packages
----------------------------------------------------------------
.. codeauthor::
    Peter Kraus

"""

import importlib
import logging

from typing import Union, TypeVar
from tomato.driverinterface_1_0 import ModelInterface as MI_1_0
from tomato.driverinterface_2_0 import ModelInterface as MI_2_0
from tomato.driverinterface_2_1 import ModelInterface as MI_2_1

ModelInterface = TypeVar("ModelInterface", MI_1_0, MI_2_0, MI_2_1)

logger = logging.getLogger(__name__)


def driver_to_interface(drivername: str) -> Union[None, ModelInterface]:
    modname = f"tomato_{drivername.replace('-', '_')}"

    try:
        mod = importlib.import_module(modname)
    except ModuleNotFoundError as e:
        logger.critical("Error when loading 'DriverInteface': %s", e)
        return None
    else:
        if hasattr(mod, "DriverInterface"):
            return getattr(mod, "DriverInterface")
        else:
            return None



import os
from pathlib import Path
import zmq
import subprocess

from tomato import tomato
from .utils import wait_until_tomato_running, wait_until_tomato_stopped

PORT = 12345
CTXT = zmq.Context()
timeout = 1000
kwargs = dict(port=PORT, context=CTXT, timeout=timeout)


def test_tomato_status_down():
    ret = tomato.status(**kwargs)
    print(f"{ret=}")
    assert ret.success is False
    assert "tomato not running" in ret.msg


def test_tomato_status_up(start_tomato_daemon, stop_tomato_daemon):
    ret = tomato.status(**kwargs)
    print(f"{ret=}")
    assert ret.success
    assert len(ret.data.pips) == 1


def test_tomato_start_no_init(datadir, stop_tomato_daemon):
    os.chdir(datadir)
    ret = tomato.start(**kwargs, appdir=Path(), verbosity=0)
    print(f"{ret=}")
    assert ret.success is False
    assert "settings file not found" in ret.msg


def test_tomato_start_with_init(datadir, stop_tomato_daemon):
    os.chdir(datadir)
    ret = tomato.init(appdir=Path(), datadir=Path(), logdir=Path())
    assert ret.success
    ret = tomato.start(**kwargs, appdir=Path(), verbosity=0)
    print(f"{ret=}")
    assert ret.success


def test_tomato_start_double(datadir, stop_tomato_daemon):
    test_tomato_start_with_init(datadir, stop_tomato_daemon)
    assert wait_until_tomato_running(port=PORT, timeout=5000)
    ret = tomato.start(**kwargs, appdir=Path(), verbosity=0)
    print(f"{ret=}")
    assert ret.success is False
    assert (
        f"port {PORT} is already in use" in ret.msg
        or f"already running on port {PORT}" in ret.msg
    )


def test_tomato_pipeline(datadir, stop_tomato_daemon):
    test_tomato_start_with_init(datadir, stop_tomato_daemon)
    ret = tomato.pipeline_load(**kwargs, pipeline="pip-counter", sampleid="test")
    print(f"{ret=}")
    assert ret.success
    assert ret.data.sampleid == "test"
    assert ret.data.ready is False

    ret = tomato.pipeline_load(**kwargs, pipeline="pip-counter", sampleid="abcdefg")
    print(f"{ret=}")
    assert ret.success is False
    assert "pipeline 'pip-counter' is not empty" in ret.msg
    assert ret.data.sampleid == "test"

    ret = tomato.pipeline_ready(**kwargs, pipeline="pip-counter")
    print(f"{ret=}")
    assert ret.success
    assert ret.data.sampleid == "test"
    assert ret.data.ready

    ret = tomato.pipeline_ready(**kwargs, pipeline="pip-counter")
    print(f"{ret=}")
    assert ret.success
    assert ret.data.sampleid == "test"
    assert ret.data.ready

    ret = tomato.pipeline_eject(**kwargs, pipeline="pip-counter")
    print(f"{ret=}")
    assert ret.success
    assert ret.data.sampleid is None
    assert ret.data.ready is False

    ret = tomato.pipeline_eject(**kwargs, pipeline="pip-counter")
    print(f"{ret=}")
    assert ret.success
    assert ret.data.sampleid is None
    assert ret.data.ready is False


def test_tomato_pipeline_invalid(datadir, stop_tomato_daemon):
    test_tomato_start_with_init(datadir, stop_tomato_daemon)
    ret = tomato.pipeline_load(**kwargs, pipeline="bogus", sampleid="test")
    print(f"{ret=}")
    assert ret.success is False
    assert "pipeline 'bogus' not found" in ret.msg

    ret = tomato.pipeline_eject(**kwargs, pipeline="bogus")
    print(f"{ret=}")
    assert ret.success is False
    assert "pipeline 'bogus' not found" in ret.msg

    ret = tomato.pipeline_ready(**kwargs, pipeline="bogus")
    print(f"{ret=}")
    assert ret.success is False
    assert "pipeline 'bogus' not found" in ret.msg


def test_tomato_log_verbosity_0(datadir, stop_tomato_daemon):
    test_tomato_start_with_init(datadir, stop_tomato_daemon)
    assert Path("daemon_12345.log").exists()
    assert Path("daemon_12345.log").stat().st_size > 0


def test_tomato_log_verbosity_testing(datadir, start_tomato_daemon, stop_tomato_daemon):
    assert wait_until_tomato_running(port=PORT, timeout=5000)
    assert Path("daemon_12345.log").exists()
    assert Path("daemon_12345.log").stat().st_size > 0


def test_tomato_log_verbosity_default(datadir, stop_tomato_daemon):
    os.chdir(datadir)
    subprocess.run(["tomato", "init", "-p", f"{PORT}", "-A", ".", "-D", ".", "-L", "."])
    subprocess.run(["tomato", "start", "-p", f"{PORT}", "-A", "."])
    assert wait_until_tomato_running(port=PORT, timeout=5000)
    assert Path("daemon_12345.log").exists()
    assert Path("daemon_12345.log").stat().st_size > 0


def test_tomato_nocmd(start_tomato_daemon, stop_tomato_daemon):
    assert wait_until_tomato_running(port=PORT, timeout=5000)
    req = CTXT.socket(zmq.REQ)
    req.connect("tcp://127.0.0.1:12345")
    req.send_pyobj(dict(cdm="typo"))
    rep = req.recv_pyobj()
    print(f"{rep=}")
    assert rep.success is False
    assert "msg without cmd" in rep.msg


def test_tomato_stop(start_tomato_daemon, stop_tomato_daemon):
    assert wait_until_tomato_running(port=PORT, timeout=5000)
    ret = tomato.stop(**kwargs)
    assert ret.success
    assert wait_until_tomato_stopped(port=PORT, timeout=5000)

    assert Path("daemon_12345.log").exists()
    with Path("daemon_12345.log").open() as logf:
        text = logf.read()
    assert "driver manager thread joined" in text
    assert "job manager thread joined" in text



import pytest
import os
import subprocess
import xarray as xr
from datetime import datetime

from tomato.models import Job
from . import utils

PORT = 12345


@pytest.mark.parametrize(
    "case, nreps",
    [
        ("counter_stresstest", 5),
    ],
)
def test_stresstest(case, nreps, datadir, stop_tomato_daemon):
    os.chdir(datadir)
    subprocess.run(["tomato", "init", "-p", f"{PORT}", "-A", ".", "-D", ".", "-L", "."])
    subprocess.run(["tomato", "start", "-p", f"{PORT}", "-A", "."])
    utils.wait_until_tomato_running(port=PORT, timeout=3000)

    subprocess.run(["tomato", "pipeline", "load", "-p", f"{PORT}", "pip-counter", case])
    for i in range(nreps):
        subprocess.run(["ketchup", "submit", "-p", f"{PORT}", f"{case}.yml"])

    subprocess.run(["tomato", "pipeline", "ready", "-p", f"{PORT}", "pip-counter"])

    utils.wait_until_ketchup_status(jobid=nreps, status="c", port=PORT, timeout=40000)

    prev = None
    for i in range(nreps):
        i += 1
        assert os.path.exists(f"results.{i}.nc")
        dt = xr.open_datatree(f"results.{i}.nc")
        completed_at = Job.model_validate_json(dt.attrs["tomato_Job"]).completed_at
        ti = datetime.fromisoformat(completed_at)
        if prev is not None:
            assert ti > prev
        prev = ti



"""
**tomato.daemon.cmd**: command parsing for tomato daemon
--------------------------------------------------------
.. codeauthor::
    Peter Kraus

All functions in this module expect a :class:`dict` containing the command specification
and a :class:`~tomato.models.Daemon` object as arguments. The :class:`Daemon` object is
altered by the command.

All functions in this module return a :class:`~tomato.models.Reply`.

"""

from tomato.models import (
    Daemon,
    Driver,
    Device,
    Reply,
    Pipeline,
    Job,
    Component,
)
from pydantic import BaseModel
from typing import Any
import logging

import tomato.daemon.io as io
import tomato.daemon.jobdb as jobdb

logger = logging.getLogger(__name__)


def merge_pipelines(
    cur: dict[str, Pipeline], new: dict[str, Pipeline]
) -> dict[str, Pipeline]:
    """
    Helper function for merging a :class:`dict` of new :class:`Pipelines` into the
    current :class:`dict`.
    """
    ret = {}
    for pname, pip in cur.items():
        if pname not in new:
            if pip.jobid is not None:
                ret[pname] = pip
        else:
            if pip.devs == new[pname].devs:
                ret[pname] = pip
            elif pip.jobid is None:
                ret[pname] = new[pname]
    for pname, pip in new.items():
        if pname not in cur:
            ret[pname] = pip
    return ret


def status(msg: dict, daemon: Daemon) -> Reply:
    return Reply(success=True, msg=daemon.status, data=daemon)


def stop(msg: dict, daemon: Daemon) -> Reply:
    logger = logging.getLogger(f"{__name__}.stop")
    logger.debug("%s", msg)
    io.store(daemon)
    if any([pip.jobid is not None for pip in daemon.pips.values()]):
        logger.error("cannot stop tomato-daemon as jobs are running")
        return Reply(success=False, msg="jobs are running")
    else:
        daemon.status = "stop"
        logger.critical("stopping tomato-daemon")
        return Reply(success=True)


def setup(msg: dict, daemon: Daemon) -> Reply:
    logger = logging.getLogger(f"{__name__}.setup")
    logger.debug("%s", msg)
    if daemon.status == "bootstrap":
        for key in ["drvs", "devs", "pips", "cmps"]:
            setattr(daemon, key, msg[key])
        logger.info("setup successful with pipelines: '%s'", daemon.pips.keys())
        daemon.status = "running"
    else:
        # First, check that we're not touching anything associated with a running job
        check_components = set()
        check_devices = set()
        check_drivers = set()
        for dpip in daemon.pips.values():
            if dpip.jobid is None:
                continue
            if dpip.name not in msg["pips"]:
                return Reply(
                    success=False,
                    msg="reload would delete a running pipeline",
                    data=dpip,
                )
            pip = msg["pips"][dpip.name]
            if pip.components != dpip.components:
                return Reply(
                    success=False,
                    msg="reload would modify components of a running pipeline",
                    data=dpip,
                )
            check_components.update(dpip.components)

        for cname in check_components:
            dcomp = daemon.cmps[cname]
            if cname not in msg["cmps"]:
                return Reply(
                    success=False,
                    msg="reload would delete a component of a running pipeline",
                    data=dcomp,
                )
            comp = msg["cmps"][cname]
            if (
                dcomp.name != comp.name
                or dcomp.driver != comp.driver
                or dcomp.device != comp.device
                or dcomp.address != comp.address
                or dcomp.channel != comp.channel
                or dcomp.role != comp.role
            ):
                return Reply(
                    success=False,
                    msg="reload would modify a component of a running pipeline",
                    data=dcomp,
                )
            check_devices.add(dcomp.device)
            check_drivers.add(dcomp.driver)

        for dname in check_devices:
            ddev = daemon.devs[dname]
            if dname not in msg["devs"]:
                return Reply(
                    success=False,
                    msg="reload would delete a device of a component in a running pipeline",
                    data=ddev,
                )
            dev = msg["devs"][dname]
            if (
                ddev.name != dev.name
                or ddev.driver != dev.driver
                or ddev.address != dev.address
                or ddev.pollrate != dev.pollrate
                or any(ch not in dev.channels for ch in ddev.channels)
            ):
                return Reply(
                    success=False,
                    msg="reload would modify a device of a component in a running pipeline",
                    data=ddev,
                )

        for dname in check_drivers:
            ddrv = daemon.drvs[dname]
            if dname not in msg["drvs"]:
                return Reply(
                    success=False,
                    msg="reload would delete a driver of a device in a running pipeline",
                    data=ddev,
                )
            drv = msg["drvs"][dname]
            if ddrv.name != drv.name or ddrv.settings != drv.settings:
                return Reply(
                    success=False,
                    msg="reload would modify a driver of a device in a running pipeline",
                    data=ddrv,
                )

        _api_reload(msg["drvs"], daemon.drvs, "driver", ["settings"])

        _api_reload(msg["pips"], daemon.pips, "pipeline", ["components"])

        attrlist = ["driver", "device", "address", "channel", "role"]
        _api_reload(msg["cmps"], daemon.cmps, "component", attrlist)

        _api_reload(msg["devs"], daemon.devs, "device", ["channels", "pollrate"])

        logger.info("reload successful with pipelines: '%s'", daemon.pips.keys())
    return Reply(success=True, data=daemon)


def _api_reload(mdict: dict, ddict: dict, objname: str, attrlist: list[str]):
    for obj in mdict.values():
        if obj.name not in ddict:
            logger.debug("adding new %s '%s'", objname, obj.name)
            ddict[obj.name] = obj
            continue
        dobj = ddict[obj.name]
        for attr in attrlist:
            if getattr(dobj, attr) != getattr(obj, attr):
                logger.debug("%s '%s.%s' updated", objname, dobj.name, attr)
                setattr(dobj, attr, getattr(obj, attr))
    for dobj in ddict.copy().values():
        if dobj.name not in mdict:
            logger.warning("removing unused %s '%s'", objname, dobj.name)
            del ddict[dobj.name]


def pipeline(msg: dict, daemon: Daemon) -> Reply:
    logger = logging.getLogger(f"{__name__}.pipeline")
    logger.debug("%s", msg)
    pip = msg["params"]
    if pip["name"] is None:
        logger.error()
        return Reply(success=False, msg="no pipeline name supplied", data=msg)
    if pip["name"] not in daemon.pips:
        dest = Pipeline(**pip)
        daemon.pips[pip["name"]] = dest
        return Reply(success=True, msg="pipeline created", data=dest)

    dest = daemon.pips[pip["name"]]
    if pip.get("delete", False) and dest.jobid is None:
        logger.warning("deleting pipeline '%s'", dest.name)
        del daemon.pips[pip["name"]]
        return Reply(success=True, msg="pipeline deleted")

    elif pip.get("delete", False):
        logger.error("cannot delete pipeline '%s' as a job is running", dest.name)
        return Reply(success=False, msg="pipeline cannot be deleted", data=dest)
    else:
        for k, v in pip.items():
            logger.debug("setting pipeline '%s.%s' to '%s'", dest.name, k, v)
            setattr(dest, k, v)
        return Reply(success=True, msg="pipeline updated", data=dest)


def set_job(msg: dict, daemon: Daemon) -> Reply:
    logger = logging.getLogger(f"{__name__}.set_job")
    dbpath = daemon.settings["jobs"]["dbpath"]
    if msg["id"] is None:
        job = Job(**msg.get("params", {}))
        job.id = jobdb.insert_job(job, dbpath)
        logger.info("received job.id %d", job.id)
    else:
        job = jobdb.update_job_id(msg["id"], msg.get("params", {}), dbpath)
        logger.info("updated job.id %d", job.id)
    return Reply(success=True, msg="job updated", data=job)


def get_jobs(msg: dict, daemon: Daemon) -> Reply:
    dbpath = daemon.settings["jobs"]["dbpath"]
    jobs = jobdb.get_jobs_where(msg["where"], dbpath)
    return Reply(success=True, msg=f"found {len(jobs)} jobs", data=jobs)


def driver(msg: dict, daemon: Daemon) -> Reply:
    return _api(
        otype="driver",
        msg=msg,
        ddict=daemon.drvs,
        Cls=Driver,
    )


def device(msg: dict, daemon: Daemon) -> Reply:
    return _api(
        otype="device",
        msg=msg,
        ddict=daemon.devs,
        Cls=Device,
    )


def component(msg: dict, daemon: Daemon) -> Reply:
    return _api(
        otype="component",
        msg=msg,
        ddict=daemon.cmps,
        Cls=Component,
    )


def _api(otype: str, msg: dict, ddict: dict[str, Any], Cls: BaseModel) -> Reply:
    logger = logging.getLogger(f"{__name__}.{otype}")
    logger.debug("%s", msg)
    obj = msg["params"]
    if obj["name"] is None:
        logger.error("no %s name supplied", otype)
        return Reply(success=False, msg=f"no {otype} name supplied", data=msg)

    if obj["name"] not in ddict:
        ddict[obj["name"]] = Cls(**obj)
        return Reply(
            success=True,
            msg=f"{otype} {obj['name']!r} created",
            data=ddict[obj["name"]],
        )
    else:
        for k, v in obj.items():
            logger.debug("setting %s '%s.%s' to '%s'", otype, obj["name"], k, v)
            setattr(ddict[obj["name"]], k, v)
        return Reply(
            success=True,
            msg=f"{otype} {obj['name']!r} updated",
            data=ddict[obj["name"]],
        )



"""
**tomato.daemon.io**: functions for storing and loading data
------------------------------------------------------------
.. codeauthor::
    Peter Kraus

"""

import pickle
import logging
import xarray as xr
import importlib.metadata
from pathlib import Path
from tomato.models import Daemon, Job

logger = logging.getLogger(__name__)


def store(daemon: Daemon):
    """Stores the status of the provided :class:`Daemon` as a ``pickle`` file."""
    datadir = Path(daemon.settings["datadir"])
    datadir.mkdir(parents=True, exist_ok=True)
    outfile = datadir / f"tomato_state_{daemon.port}.pkl"
    logger.debug("storing daemon state to '%s'", outfile)
    with outfile.open("wb") as out:
        pickle.dump(daemon, out, protocol=5)


def load(daemon: Daemon):
    """Restores a saved status from a ``pickle`` file into the provided :class:`Daemon`."""
    infile = Path(daemon.settings["datadir"]) / f"tomato_state_{daemon.port}.pkl"
    if infile.exists() is False:
        logger.debug("daemon state file '%s' does not exist", infile)
        return
    with infile.open("rb") as inp:
        loaded = pickle.load(inp)
    daemon.pips = loaded.pips
    daemon.devs = loaded.devs
    daemon.drvs = loaded.drvs
    daemon.cmps = loaded.cmps
    daemon.status = "running"


def merge_netcdfs(job: Job, snapshot=False):
    """
    Merges the individual pickled :class:`xr.Datasets` of each Component found in :obj:`job.jobpath`
    into a single :class:`xr.DataTree`, which is then stored in the NetCDF file,
    using the Component `role` as the group label.
    """
    logger = logging.getLogger(f"{__name__}.merge_netcdf")
    logger.debug("opening datasets")
    datasets = []
    logger.debug(f"{job=}")
    logger.debug(f"{job.jobpath=}")
    for fn in Path(job.jobpath).glob("*.pkl"):
        with pickle.load(fn.open("rb")) as ds:
            datasets.append(ds)
    logger.debug("creating a DataTree from %d groups", len(datasets))
    dt = xr.DataTree.from_dict({ds.attrs["role"]: ds for ds in datasets})
    logger.debug(f"{dt=}")
    root_attrs = {
        "tomato_version": importlib.metadata.version("tomato"),
        "tomato_Job": job.model_dump_json(),
    }
    dt.attrs = root_attrs
    outpath = job.snappath if snapshot else job.respath
    logger.debug("saving DataTree into '%s'", outpath)
    dt.to_netcdf(outpath, engine="h5netcdf")
    logger.debug(f"{dt=}")


def data_to_pickle(ds: xr.Dataset, path: Path, role: str):
    """
    Dumps the data provided as :class:`xr.Dataset` into a ``pickle``. Concatenates with
    any existing data stored in the ``pickle``.
    """
    logger = logging.getLogger(f"{__name__}.data_to_pickle")
    ds.attrs["role"] = role
    logger.debug("checking for existing pickle at '%s'", path)
    if path.exists():
        with pickle.load(path.open("rb")) as oldds:
            logger.debug("concatenating Dataset with existing data")
            ds = xr.concat([oldds, ds], dim="uts")
    logger.debug("dumping Dataset into pickle at '%s'", path)
    with path.open("wb") as out:
        pickle.dump(ds, out, protocol=5)



import sys
from pathlib import Path

from importlib import metadata
import argparse
import logging
import zmq
import appdirs
import yaml

from tomato import tomato, ketchup, passata

sys.path += sys.modules["tomato"].__path__

__version__ = metadata.version("tomato")
VERSION = __version__
DEFAULT_TOMATO_PORT = 1234
logger = logging.getLogger(__name__)


def set_loglevel(loglevel: int):
    logging.basicConfig(level=loglevel)
    logger.debug("loglevel set to '%s'", logging._levelToName[loglevel])


def parse_args(parser, verbose, is_tomato=False):
    args, extras = parser.parse_known_args()
    args, extras = verbose.parse_known_args(extras, args)

    verbosity = min(max((2 + args.quiet - args.verbose) * 10, 10), 50)
    set_loglevel(verbosity)

    context = zmq.Context()
    if not is_tomato:
        status = tomato.status(**vars(args), context=context)
        if not status.success:
            if args.yaml:
                print(yaml.dump(status.dict()))
            else:
                print(f"Failure: {status.msg}")

    if "func" in args:
        ret = args.func(**vars(args), verbosity=verbosity, context=context)
        if args.yaml:
            print(yaml.dump(ret.dict()))
        else:
            print(f"{'Success' if ret.success else 'Failure'}: {ret.msg}")


def run_tomato():
    dirs = appdirs.AppDirs("tomato", "dgbowl", version=VERSION)
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s version {VERSION}",
    )

    verbose = argparse.ArgumentParser(add_help=False)

    subparsers = parser.add_subparsers(dest="subcommand", required=True)

    status = subparsers.add_parser("status")
    status.add_argument(
        "stgrp",
        choices=["tomato", "pipelines", "drivers", "devices", "components"],
        default="tomato",
        nargs="?",
        help="Check status of 'tomato', or its 'pipelines', 'drivers', 'devices', or 'components'.",
    )
    status.set_defaults(func=tomato.status)

    start = subparsers.add_parser("start")
    start.set_defaults(func=tomato.start)

    stop = subparsers.add_parser("stop")
    stop.set_defaults(func=tomato.stop)

    init = subparsers.add_parser("init")
    init.set_defaults(func=tomato.init)

    reload = subparsers.add_parser("reload")
    reload.set_defaults(func=tomato.reload)

    pipeline = subparsers.add_parser("pipeline")
    pipparsers = pipeline.add_subparsers(dest="subsubcommand", required=True)

    pip_load = pipparsers.add_parser("load")
    pip_load.set_defaults(func=tomato.pipeline_load)
    pip_load.add_argument("pipeline")
    pip_load.add_argument("sampleid")

    pip_eject = pipparsers.add_parser("eject")
    pip_eject.set_defaults(func=tomato.pipeline_eject)
    pip_eject.add_argument("pipeline")

    pip_ready = pipparsers.add_parser("ready")
    pip_ready.set_defaults(func=tomato.pipeline_ready)
    pip_ready.add_argument("pipeline")

    for p in [parser, verbose]:
        p.add_argument(
            "--verbose",
            "-v",
            action="count",
            default=0,
            help="Increase verbosity of tomato daemon by one level.",
        )
        p.add_argument(
            "--quiet",
            "-q",
            action="count",
            default=0,
            help="Decrease verbosity of tomato daemon by one level.",
        )

    for p in [start, stop, init, status, reload, pip_load, pip_eject, pip_ready]:
        p.add_argument(
            "--port",
            "-p",
            help="Port number of tomato's reply socket",
            default=DEFAULT_TOMATO_PORT,
            type=int,
        )
        p.add_argument(
            "--timeout",
            help="Timeout for the tomato command, in milliseconds",
            type=int,
            default=3000,
        )
        p.add_argument(
            "--yaml",
            "-y",
            help="Return output as a yaml.",
            action="store_true",
            default=False,
        )

    for p in [start, init, reload]:
        p.add_argument(
            "--appdir",
            "-A",
            type=Path,
            help="Settings directory for tomato",
            default=Path(dirs.user_config_dir),
        )
    for p in [init]:
        p.add_argument(
            "--datadir",
            "-D",
            type=Path,
            help="Data directory for tomato",
            default=Path(dirs.user_data_dir),
        )
        p.add_argument(
            "--logdir",
            "-L",
            type=Path,
            help="Log directory for tomato",
            default=Path(dirs.user_log_dir),
        )

    parse_args(parser, verbose, is_tomato=True)


def run_ketchup():
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s version {VERSION}",
    )

    verbose = argparse.ArgumentParser(add_help=False)

    for p in [parser, verbose]:
        p.add_argument(
            "--verbose",
            "-v",
            action="count",
            default=0,
            help="Increase verbosity by one level.",
        )
        p.add_argument(
            "--quiet",
            "-q",
            action="count",
            default=0,
            help="Decrease verbosity by one level.",
        )

    subparsers = parser.add_subparsers(dest="subcommand", required=True)

    submit = subparsers.add_parser("submit")
    submit.add_argument(
        "payload",
        help="File containing the payload to be submitted to tomato.",
        default=None,
    )
    submit.add_argument(
        "-j",
        "--jobname",
        help="Set the job name of the submitted job to?",
        default=None,
    )
    submit.set_defaults(func=ketchup.submit)

    status = subparsers.add_parser("status")
    status.add_argument(
        "jobids",
        nargs="*",
        help=(
            "The job.id(s) of the job(s) to be checked. "
            "Defaults to the status of the whole queue."
        ),
        type=int,
        default=None,
    )
    status.set_defaults(func=ketchup.status)

    cancel = subparsers.add_parser("cancel")
    cancel.add_argument(
        "jobids",
        nargs="+",
        help=(
            "The job.id(s) of the job(s) to be cancelled. "
            "At least one job.id has to be provided."
        ),
        type=int,
        default=None,
    )
    cancel.set_defaults(func=ketchup.cancel)

    snapshot = subparsers.add_parser("snapshot")
    snapshot.add_argument(
        "jobids",
        nargs="+",
        help=(
            "The job.id(s) of the job(s) to be snapshotted. "
            "At least one job.id has to be provided."
        ),
        type=int,
        default=None,
    )
    snapshot.set_defaults(func=ketchup.snapshot)

    search = subparsers.add_parser("search")
    search.add_argument(
        "jobname",
        help="The jobname of the searched job.",
        default=None,
    )
    search.add_argument(
        "-c",
        "--complete",
        action="store_true",
        default=False,
        help="Search also in completed jobs.",
    )
    search.set_defaults(func=ketchup.search)

    for p in [submit, status, cancel, snapshot, search]:
        p.add_argument(
            "--port",
            "-p",
            type=int,
            help="Port number of tomato's reply socket",
            default=DEFAULT_TOMATO_PORT,
        )
        p.add_argument(
            "--timeout",
            help="Timeout for the ketchup command, in milliseconds",
            type=int,
            default=3000,
        )
        p.add_argument(
            "--yaml",
            "-y",
            help="Return output as a yaml.",
            action="store_true",
            default=False,
        )

    parse_args(parser, verbose, is_tomato=False)


def run_passata():
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s version {VERSION}",
    )

    verbose = argparse.ArgumentParser(add_help=False)

    for p in [parser, verbose]:
        p.add_argument(
            "--verbose",
            "-v",
            action="count",
            default=0,
            help="Increase verbosity by one level.",
        )
        p.add_argument(
            "--quiet",
            "-q",
            action="count",
            default=0,
            help="Decrease verbosity by one level.",
        )

    subparsers = parser.add_subparsers(dest="subcommand", required=True)
    stats = subparsers.add_parser("status")
    attrs = subparsers.add_parser("attrs")
    regis = subparsers.add_parser("register")
    capbs = subparsers.add_parser("capabilities")
    const = subparsers.add_parser("constants")
    gattr = subparsers.add_parser("get")

    for p in [stats, attrs, regis, capbs, gattr, const]:
        p.add_argument(
            "name",
            help=(
                "The Component.name of the component to be queried. "
                "At least one Component.name has to be provided."
            ),
            type=str,
        )
        p.add_argument(
            "--port",
            "-p",
            help="Port number of tomato's reply socket",
            default=DEFAULT_TOMATO_PORT,
            type=int,
        )
        p.add_argument(
            "--timeout",
            help="Timeout for the tomato command, in milliseconds",
            type=int,
            default=3000,
        )
        p.add_argument(
            "--yaml",
            "-y",
            help="Return output as a yaml.",
            action="store_true",
            default=False,
        )

    gattr.add_argument(
        "attrs",
        help="The attribute name(s) to be queried.",
        nargs="+",
    )

    stats.set_defaults(func=passata.status)
    attrs.set_defaults(func=passata.attrs)
    regis.set_defaults(func=passata.register)
    capbs.set_defaults(func=passata.capabilities)
    const.set_defaults(func=passata.constants)
    gattr.set_defaults(func=passata.get_attrs)

    parse_args(parser, verbose, is_tomato=False)



from tomato.models import Reply
from tomato.driverinterface_2_1.types import Val

from functools import wraps
import logging
import sys
import pint

logger = logging.getLogger(__name__)


def in_devmap(func):
    @wraps(func)
    def wrapper(self, **kwargs):
        if "key" in kwargs:
            key = kwargs.pop("key")
        else:
            address = kwargs.get("address")
            channel = kwargs.get("channel")
            key = (address, channel)
        if key not in self.devmap:
            msg = f"component with key {key!r} is not registered"
            return Reply(success=False, msg=msg, data=None)
        return func(self, **kwargs, key=key)

    return wrapper


def to_reply(func):
    """
    Helper decorator for coercing tuples into :class:`Reply`.
    """

    @wraps(func)
    def wrapper(self, **kwargs):
        ret = func(self, **kwargs)
        if isinstance(ret, Reply):
            return ret
        else:
            success, msg, data = ret
            return Reply(success=success, msg=msg, data=data)

    return wrapper


def log_errors(func):
    """
    Helper decorator for logging all kinds of errors.

    This decorator should be only used on functions in the API of the
    :class:`ModelInterface`, as the caught exceptions will cause the
    driver process to exit.
    """

    @wraps(func)
    def wrapper(self, **kwargs):
        try:
            return func(self, **kwargs)
        # We want to preserve TypeErrors, ValueErrors and AttributeErrors for testing.
        # These should be then caught in the tomato-driver process.
        except (ValueError, AttributeError) as e:
            logger.critical(e, exc_info=True)
            raise e
        # Other kinds of errors we abort the driver process
        except Exception as e:
            logger.critical(e, exc_info=True)
            sys.exit(e)

    return wrapper


def coerce_val(func):
    """
    Decorator for coercing :obj:`val` into the correct format based on :class:`Attr` data.

    This decorator should be applied to the :func:`ModelDriver.set_attr` function, in
    order to check whether the supplied value is allowed (not ``None``, in ``options``,
    between ``minimum`` and ``maximum``) as well as coercing it to the right type and
    unit.
    """

    @wraps(func)
    def wrapper(self, attr: str, val: Val, **kwargs: dict) -> Val:
        if val is None:
            raise ValueError(f"attr {attr!r} cannot be None")
        if attr not in self.attrs():
            raise AttributeError(f"unknown attr: {attr!r}")
        props = self.attrs()[attr]
        if not props.rw:
            raise AttributeError(f"attr {attr!r} is read-only")

        if not isinstance(val, props.type):
            # This may raise ValueError
            val = props.type(val)
        if props.options is not None and val not in props.options:
            raise ValueError(f"val {val!r} is not in allowed options {props.options}")

        if isinstance(val, pint.Quantity):
            if val.dimensionless and props.units is not None:
                val = pint.Quantity(val.m, props.units)
            if val.dimensionality != pint.Quantity(props.units).dimensionality:
                raise ValueError(f"val {val!r} has the wrong dimensionality")
        if props.minimum is not None and val < props.minimum:
            raise ValueError(f"val {val!r} is smaller than {props.minimum}")
        if props.maximum is not None and val > props.maximum:
            raise ValueError(f"val {val!r} is greater than {props.maximum}")

        return func(self, attr=attr, val=val, **kwargs)

    return wrapper



"""
**tomato.daemon.jobdb**: the sqlite database for jobs in tomato
---------------------------------------------------------------
.. codeauthor::
    Peter Kraus

"""

import sqlite3
import logging
import os
import pickle
from tomato.models import Job

logger = logging.getLogger(__name__)


def connect_jobdb(dbpath: str):
    head, tail = os.path.split(dbpath)
    if head != "" and not os.path.exists(head):
        logger.warning("making local data folder '%s'", head)
        os.makedirs(head)
    conn = sqlite3.connect(dbpath)
    cur = conn.cursor()
    return conn, cur


def jobdb_setup(dbpath: str) -> None:
    user_version = 1
    conn, cur = connect_jobdb(dbpath)
    logger.debug("attempting to find table 'queue' in '%s'", dbpath)
    cur.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='queue';")
    exists = bool(len(cur.fetchall()))
    if exists:
        logger.debug("table 'queue' present at '%s'", dbpath)
        cur.execute("PRAGMA user_version;")
        curr_version = cur.fetchone()[0]
        assert curr_version == user_version
        # Below is an example of upgrading databases to new user_version:
        # while curr_version < user_version:
        #    if curr_version == 0:
        #        log.info("upgrading table 'queue' from version 0 to 1")
        #        cur.execute("ALTER TABLE queue ADD COLUMN jobname TEXT;")
        #        cur.execute("PRAGMA user_version = 1;")
        #        conn.commit()
        #    cur.execute("PRAGMA user_version;")
        #    curr_version = cur.fetchone()[0]
    else:
        logger.info("creating a new sqlite3 'queue' table at '%s'", dbpath)
        cur.execute(
            "CREATE TABLE IF NOT EXISTS queue ("
            "    id INTEGER PRIMARY KEY AUTOINCREMENT,"
            "    payload BLOB NOT NULL,"
            "    jobname TEXT,"
            "    pid INTEGER,"
            "    status TEXT NOT NULL,"
            "    submitted_at TEXT NOT NULL,"
            "    executed_at TEXT,"
            "    completed_at TEXT,"
            "    jobpath TEXT,"
            "    respath TEXT,"
            "    snappath TEXT"
            ");",
        )
        cur.execute(f"PRAGMA user_version = {user_version};")
        conn.commit()
    conn.close()


def insert_job(job: Job, dbpath: str) -> int:
    conn, cur = connect_jobdb(dbpath)
    cur.execute(
        "INSERT INTO queue (payload, jobname, pid, status, submitted_at, "
        "executed_at, completed_at, jobpath, respath, snappath)"
        "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?);",
        (
            pickle.dumps(job.payload),
            job.jobname,
            job.pid,
            job.status,
            job.submitted_at,
            job.executed_at,
            job.completed_at,
            job.jobpath,
            job.respath,
            job.snappath,
        ),
    )
    conn.commit()
    cur.execute(f"SELECT id FROM queue WHERE submitted_at = '{job.submitted_at}';")
    ret = cur.fetchone()[0]
    conn.close()
    return ret


def update_job_id(id: int, params: dict, dbpath: str) -> Job:
    conn, cur = connect_jobdb(dbpath)
    for k, v in params.items():
        cur.execute(f"UPDATE queue SET {k} = ? WHERE id = {id};", (v,))
    conn.commit()
    conn.close()
    return get_job_id(id, dbpath)


def get_job_id(id: int, dbpath: str) -> Job:
    conn, cur = connect_jobdb(dbpath)
    cur.execute("SELECT * FROM queue WHERE id = ?;", (id,))
    columns = [i[0] for i in cur.description]
    data = cur.fetchone()
    conn.close()
    j = Job(**{k: v for k, v in zip(columns, data)})
    return j


def get_jobs_where(where: str, dbpath: str) -> list[Job]:
    conn, cur = connect_jobdb(dbpath)
    cur.execute(f"SELECT * FROM queue WHERE {where};")
    columns = [i[0] for i in cur.description]
    data = cur.fetchall()
    conn.close()
    jobs = []
    for row in data:
        jobs.append(Job(**{k: v for k, v in zip(columns, row)}))
    return jobs



"""
**tomato.daemon.driver**: the driver manager of tomato daemon
-------------------------------------------------------------
.. codeauthor::
    Peter Kraus

"""

import os
import subprocess
import logging
import time
import argparse
from importlib import metadata
from collections import defaultdict
from datetime import datetime, timezone
from threading import current_thread
from pathlib import Path
from typing import Union, TypeVar

import zmq
import psutil

from tomato.driverinterface_1_0 import ModelInterface as MI_1_0
from tomato.driverinterface_2_0 import ModelInterface as MI_2_0
from tomato.driverinterface_2_1 import ModelInterface as MI_2_1
from tomato.drivers import driver_to_interface
from tomato.models import Reply, Daemon

logger = logging.getLogger(__name__)
ModelInterface = TypeVar("ModelInterface", MI_1_0, MI_2_0, MI_2_1)
IDLE_MEASUREMENT_INTERVAL = None
MAX_REGISTER_RETRIES = 3


def tomato_driver_bootstrap(
    req: zmq.Socket, logger: logging.Logger, interface: ModelInterface, driver: str
):
    logger.debug("getting daemon status")
    req.send_pyobj(dict(cmd="status"))
    daemon: Daemon = req.recv_pyobj().data
    drv = daemon.drvs[driver]
    interface.settings = drv.settings

    logger.info("registering components for driver '%s'", driver)
    for comp in daemon.cmps.values():
        if comp.driver == driver:
            key = (comp.address, comp.channel)
            if interface.retries.get(key, 0) == MAX_REGISTER_RETRIES:
                logger.warning(
                    "component %s has exceeded MAX_REGISTER_RETRIES, skipping",
                    comp.name,
                )
                continue
            logger.info("registering component %s", comp.name)
            ret = interface.dev_register(address=comp.address, channel=comp.channel)
            if ret.success:
                logger.debug("registered component %s: %s", comp.name, ret.msg)
            else:
                logger.critical(
                    "failed to register component %s: %s", comp.name, ret.msg
                )
            params = dict(name=comp.name, capabilities=ret.data)
            req.send_pyobj(dict(cmd="component", params=params))
            ret = req.recv_pyobj()
    logger.info("driver '%s' bootstrapped successfully", driver)


def perform_idle_measurements(
    interface: ModelInterface, t_last: Union[float, None]
) -> Union[float, None]:
    if not hasattr(interface, "cmp_measure"):
        return t_last

    if "idle_measurement_interval" in interface.settings:
        imi = interface.settings["idle_measurement_interval"]
    elif hasattr(interface, "idle_measurement_interval"):
        imi = interface.idle_measurement_interval
    else:
        imi = IDLE_MEASUREMENT_INTERVAL
    if imi is None:
        return None

    t_now = time.perf_counter()
    if t_last is not None and t_now - t_last < imi:
        return t_last
    for key in interface.devmap.keys():
        interface.cmp_measure(key=key)
    return t_now


def tomato_driver() -> None:
    """
    The function called when `tomato-driver` is executed.

    This function is responsible for managing all activities involving devices of a
    single driver type.

    First, the list of devices (and their channel/address) for the specified driver is
    fetched from the `tomato-daemon`. Then, a new instance of the specified driver is
    spawned, populating its device map using the above list. If successful, the current
    process information is fed back to the `tomato-daemon`.

    Afterwards, the main loop handles all requests related to each of the devices
    managed by this driver process, including job commands. Finally, if the driver is
    instructed to stop, it attempts to perform a teardown before exiting.
    """
    # ARGUMENT PARSING
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s version {metadata.version('tomato')}",
    )
    parser.add_argument(
        "--port",
        help="Port of the tomato-daemon.",
        default=1234,
        type=int,
    )
    parser.add_argument(
        "--verbosity",
        help="Verbosity of the tomato-driver.",
        default=logging.DEBUG,
        type=int,
    )
    parser.add_argument(
        "--logdir",
        help="Logging directory for the tomato-driver.",
        default=".",
        type=str,
    )
    parser.add_argument(
        "driver",
        type=str,
        help="Name of the driver module.",
    )
    args = parser.parse_args()

    # LOGFILE
    logfile = f"driver_{args.driver}_{args.port}.log"
    logpath = Path(args.logdir) / logfile
    logger = logging.getLogger(f"{__name__}.tomato_driver({args.driver!r})")
    logging.basicConfig(
        level=args.verbosity,
        format="%(asctime)s - %(levelname)8s - %(name)-30s - %(message)s",
        handlers=[logging.FileHandler(logpath, mode="a")],
    )

    # PORTS
    context = zmq.Context()
    rep = context.socket(zmq.REP)
    port = rep.bind_to_random_port("tcp://127.0.0.1")
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{args.port}")

    logger.debug("getting pid")
    if psutil.WINDOWS:
        pid = os.getppid()
    elif psutil.POSIX:
        pid = os.getpid()

    logger.info("attempting to create Interface for driver '%s'", args.driver)
    Interface = driver_to_interface(args.driver)
    if Interface is None:
        logger.critical("class DriverInterface driver '%s' not found", args.driver)
        return

    interface: ModelInterface = Interface()
    tomato_driver_bootstrap(req, logger, interface, args.driver)

    params = dict(
        name=args.driver,
        port=port,
        pid=pid,
        connected_at=str(datetime.now(timezone.utc)),
        version=interface.version,
        settings=interface.settings,
    )
    req.send_pyobj(
        dict(cmd="driver", params=params, sender=f"{__name__}.tomato_driver")
    )
    ret = req.recv_pyobj()
    if not ret.success:
        logger.error("could not push driver '%s' state to tomato-daemon", args.driver)
        logger.debug(f"{ret=}")
        return

    logger.info("driver '%s' is entering main loop", args.driver)

    poller = zmq.Poller()
    poller.register(rep, zmq.POLLIN)
    status = "running"
    t_last = None
    try:
        while True:
            socks = dict(poller.poll(100))
            if rep in socks:
                msg = rep.recv_pyobj()
                logger.debug("received msg=%s", msg)
                if "cmd" not in msg:
                    logger.error(f"received msg without cmd: {msg=}")
                    ret = Reply(success=False, msg="received msg without cmd", data=msg)
                elif msg["cmd"] == "status":
                    ret = Reply(
                        success=True,
                        msg=f"status of driver {params['name']!r} is {status!r}",
                        data=dict(**params, status=status),
                    )
                elif msg["cmd"] == "register":
                    tomato_driver_bootstrap(req, logger, interface, args.driver)
                    if any([retry for retry in interface.retries.values()]):
                        ret = Reply(
                            success=False,
                            msg="some components not registered successfully",
                            data=interface.retries,
                        )
                    else:
                        ret = Reply(
                            success=True,
                            msg="all components re-registered successfully",
                            data=interface.retries,
                        )
                elif msg["cmd"] == "stop":
                    status = "stop"
                    ret = Reply(
                        success=True,
                        msg=f"stopping driver {args.driver!r}",
                        data=dict(status=status, driver=args.driver),
                    )
                elif msg["cmd"] == "settings":
                    interface.settings = msg["params"]
                    params["settings"] = interface.settings
                    ret = Reply(
                        success=True,
                        msg="settings received",
                        data=msg.get("params"),
                    )
                elif msg["cmd"] == "cmp_register":
                    ret = interface.cmp_register(**msg["params"])
                    cname = f"{args.driver}:({msg['params']['address']}:{msg['params']['channel']})"
                    if ret.success:
                        params = dict(name=cname, capabilities=ret.data)
                        req.send_pyobj(dict(cmd="component", params=params))
                        ret = req.recv_pyobj()
                elif hasattr(interface, msg["cmd"]):
                    try:
                        ret = getattr(interface, msg["cmd"])(**msg["params"])
                    except (ValueError, AttributeError) as e:
                        logger.info("above error caught by driver process")
                        ret = Reply(
                            success=False,
                            msg=f"{type(e)}: {str(e)}",
                            data=None,
                        )
                else:
                    logger.critical("unknown command: '%s'", msg["cmd"])
                    ret = Reply(
                        success=False,
                        msg=f"unknown command: {msg['cmd']}",
                        data=None,
                    )
                logger.debug("replying %s", ret)
                rep.send_pyobj(ret)
            if status == "stop":
                break
            elif status == "running":
                t_last = perform_idle_measurements(interface, t_last)
    except Exception as e:
        logger.critical("uncaught exception %s", type(e), exc_info=True)
        raise e

    logger.info("driver '%s' is beginning reset", args.driver)
    interface.reset()

    logger.info("driver '%s' is quitting", args.driver)


def spawn_tomato_driver(
    port: int, driver: str, req: zmq.Socket, verbosity: int, logpath: str
):
    cmd = [
        "tomato-driver",
        "--port",
        str(port),
        "--verbosity",
        str(verbosity),
        "--logdir",
        str(logpath),
        driver,
    ]
    if psutil.WINDOWS:
        cfs = subprocess.CREATE_NO_WINDOW
        cfs |= subprocess.CREATE_NEW_PROCESS_GROUP
        subprocess.Popen(cmd, creationflags=cfs)
    elif psutil.POSIX:
        subprocess.Popen(cmd, start_new_session=True)
    params = dict(name=driver, spawned_at=str(datetime.now(timezone.utc)))
    req.send_pyobj(
        dict(
            cmd="driver",
            params=params,
            sender=f"{__name__}.spawn_tomato_driver",
        )
    )
    ret = req.recv_pyobj()
    if ret.success:
        logger.info("driver process for driver '%s' launched", driver)
    else:
        logger.error("could not launch driver process for driver '%s'", driver)


def stop_tomato_driver(port: int, context):
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{port}")
    req.send_pyobj(dict(cmd="stop", sender=f"{__name__}.stop_tomato_driver"))
    return req.recv_pyobj()


def manager(port: int, timeout: int = 1000):
    """
    The driver manager thread of `tomato-daemon`.

    This manager ensures individual driver processes are (re-)spawned and instructed to
    quit as necessary.
    """
    sender = f"{__name__}.manager"
    context = zmq.Context()
    logger = logging.getLogger(sender)
    thread = current_thread()
    logger.info("launched successfully")
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{port}")
    poller = zmq.Poller()
    poller.register(req, zmq.POLLIN)
    to = timeout

    spawned_drivers = dict()
    retries = defaultdict(int)

    while getattr(thread, "do_run"):
        req.send_pyobj(dict(cmd="status", sender=sender))
        events = dict(poller.poll(to))
        if req not in events:
            logger.warning("could not contact tomato-daemon in %d ms", to)
            to = to * 2
            continue
        elif to > timeout:
            to = timeout

        daemon = req.recv_pyobj().data
        for driver in daemon.drvs.keys():
            args = [
                daemon.port,
                driver,
                req,
                daemon.verbosity,
                daemon.settings["logdir"],
            ]
            if driver not in daemon.drvs:
                logger.debug("spawning driver '%s'", driver)
                spawn_tomato_driver(*args)
                spawned_drivers[driver] = 1
            elif driver not in spawned_drivers or spawned_drivers[driver] > 5:
                drv = daemon.drvs[driver]
                if drv.pid is not None and (
                    not psutil.pid_exists(drv.pid)
                    or psutil.Process(drv.pid).status() in psutil.STATUS_ZOMBIE
                ):
                    logger.warning("respawning crashed driver '%s'", driver)
                    spawn_tomato_driver(*args)
                    spawned_drivers[driver] = 1
                elif drv.pid is None and drv.spawned_at is None:
                    logger.debug("spawning driver '%s'", driver)
                    spawn_tomato_driver(*args)
                    spawned_drivers[driver] = 1
                elif driver in spawned_drivers:
                    logger.info("driver '%s' spawned at pid %d", driver, drv.pid)
                    spawned_drivers.pop(driver)
            elif driver in spawned_drivers:
                spawned_drivers[driver] += 1

        if len(spawned_drivers) == 0:
            contact_drivers = set()
            for comp in daemon.cmps.values():
                if comp.capabilities is None:
                    key = (comp.address, comp.channel)
                    if retries[key] < MAX_REGISTER_RETRIES:
                        contact_drivers.add(comp.driver)
            for driver in contact_drivers:
                drv = daemon.drvs[driver]
                if drv.port is None:
                    continue
                logger.info("contacting driver '%s' to re-register components", driver)
                dreq = context.socket(zmq.REQ)
                dreq.connect(f"tcp://127.0.0.1:{drv.port}")
                dreq.send_pyobj(dict(cmd="register", params=None, sender=sender))
                ret = dreq.recv_pyobj()
                retries = ret.data
                dreq.close()
        time.sleep(1 if len(spawned_drivers) > 0 else 0.1)

    logger.info("instructed to quit")
    req.send_pyobj(dict(cmd="status", sender=sender))
    daemon = req.recv_pyobj().data
    for driver in daemon.drvs.values():
        logger.debug("stopping driver '%s' on port %d", driver.name, driver.port)
        ret = stop_tomato_driver(driver.port, context)
        if ret.success:
            logger.info("stopped driver '%s'", driver.name)
        else:
            logger.warning("could not stop driver '%s'", driver.name)



import pytest
import os
import subprocess
import json
import yaml
import xarray as xr
import zmq
from tomato import tomato


from . import utils

PORT = 12345
CTXT = zmq.Context()


@pytest.mark.parametrize(
    "casename, npoints",
    [
        ("psutil_1_0.1", {"psutil": 10}),
        ("psutil_counter", {"psutil": 12, "counter": 10}),
        ("psutil_counter_with_task_start", {"psutil": 12, "counter": 10}),
        ("psutil_counter_with_task_stop", {"psutil": 10, "counter": 2}),
    ],
)
def test_psutil_multidev(casename, npoints, datadir, stop_tomato_daemon):
    os.chdir(datadir)
    with open("devices_psutil.json", "r") as inf:
        jsdata = json.load(inf)
    with open("devices.yml", "w") as ouf:
        yaml.dump(jsdata, ouf)
    subprocess.run(["tomato", "init", "-p", f"{PORT}", "-A", ".", "-D", ".", "-L", "."])
    subprocess.run(["tomato", "start", "-p", f"{PORT}", "-A", ".", "-vv"])
    utils.wait_until_tomato_running(port=PORT, timeout=3000)

    utils.run_casenames([casename], [None], ["pip-multidev"])
    utils.wait_until_ketchup_status(jobid=1, status="r", port=PORT, timeout=2000)
    utils.wait_until_ketchup_status(jobid=1, status="c", port=PORT, timeout=5000)

    status = utils.job_status(1)
    assert status == "c"
    files = os.listdir(os.path.join(".", "Jobs", "1"))
    assert "jobdata.json" in files
    assert "job-1.log" in files
    assert os.path.exists("results.1.nc")
    dt = xr.open_datatree("results.1.nc")
    for group, points in npoints.items():
        print(f"{dt[group]=}")
        assert dt[group]["uts"].size == points


def test_psutil_passata(datadir, stop_tomato_daemon):
    os.chdir(datadir)
    with open("devices_psutil.json", "r") as inf:
        jsdata = json.load(inf)
    with open("devices.yml", "w") as ouf:
        yaml.dump(jsdata, ouf)
    subprocess.run(["tomato", "init", "-p", f"{PORT}", "-A", ".", "-D", ".", "-L", "."])
    subprocess.run(["tomato", "start", "-p", f"{PORT}", "-A", ".", "-vv"])
    utils.wait_until_tomato_running(port=PORT, timeout=3000)
    utils.wait_until_tomato_drivers(port=PORT, timeout=3000)

    ret = tomato.status(port=PORT, timeout=1000, context=CTXT, stgrp="drivers")
    assert ret.success
    assert ret.data["psutil"].version == "1.0"

    ret = subprocess.run(
        ["passata", "status", "psutil:(psutil-addr,10)", "-p", f"{PORT}"],
        capture_output=True,
        text=True,
    )
    print(f"{ret=}")
    assert "Success: component ('psutil-addr', '10') is not running" in ret.stdout

    ret = subprocess.run(
        ["passata", "attrs", "psutil:(psutil-addr,10)", "-p", f"{PORT}"],
        capture_output=True,
        text=True,
    )
    print(f"{ret=}")
    assert "Success: attrs of component ('psutil-addr', '10') are" in ret.stdout

    ret = subprocess.run(
        ["passata", "constants", "psutil:(psutil-addr,10)", "-p", f"{PORT}"],
        capture_output=True,
        text=True,
    )
    print(f"{ret=}")
    assert (
        "Failure: driver of component 'psutil:(psutil-addr,10)' is on version 1.0"
        in ret.stdout
    )
